import os
from abc import ABCMeta
from fnmatch import filter as fnfilter
from typing import Optional, Mapping, Union

from hbutils.string import truncate

from ..base import _process_environ
from ...control.model import Identification, ResourceLimit
from ...utils import get_repr_info


class _IGlobalConfig(metaclass=ABCMeta):
    def __init__(self, identification, resources, environ, use_sys_env):
        """
        :param identification: identification
        :param resources: resource limits
        :param environ: environment variable
        :param use_sys_env: use environment variables from local environ
        """
        self.__identification = identification
        self.__resources = resources
        self.__environ = environ
        self.__use_sys_env = use_sys_env

    def __repr__(self):
        """
        :return: get representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('identification',
                 lambda: truncate(repr(self.__identification), width=48, show_length=True, tail_length=16),
                 lambda: self.__identification and self.__identification != Identification.loads({})),
                ('resources', lambda: truncate(repr(self.__resources), width=64, show_length=True, tail_length=16),
                 lambda: self.__resources and self.__resources != ResourceLimit.loads({})),
                ('environ',
                 lambda: truncate(repr(self.__environ), width=64, show_length=True, tail_length=16),
                 lambda: self.__environ),
                ('use_sys_env', lambda: truncate(repr(self.__use_sys_env), width=64, show_length=True, tail_length=16),
                 lambda: self.__use_sys_env is not None),
            ]
        )


def _process_use_sys_env(use_sys_env) -> Union[set, bool]:
    if isinstance(use_sys_env, (list, tuple, set)):
        return set(use_sys_env)
    elif isinstance(use_sys_env, bool) or use_sys_env is None:
        return not not use_sys_env
    else:
        raise TypeError(
            'Bool or list expected but {actual} found for use_sys_env.'.format(actual=repr(type(use_sys_env).__name__)))


def _load_local_environ(use_sys_env) -> Mapping[str, str]:
    use_sys_env = _process_use_sys_env(use_sys_env)
    _current_env = dict(os.environ)
    if isinstance(use_sys_env, set):
        _keys = set()
        for pattern in use_sys_env:
            _keys |= set(fnfilter(list(_current_env.keys()), pattern))
        return {key: value for key, value in _current_env.items() if key in _keys}
    else:
        return _current_env if use_sys_env else {}


class GlobalConfigTemplate(_IGlobalConfig):
    def __init__(self, identification=None, resources=None, environ=None, use_sys_env=None):
        """
        :param identification: identification
        :param resources: resource limits
        :param environ: environment variable
        :param use_sys_env: use environment variables from local environ
        """
        self.__identification = Identification.loads(identification)
        self.__resources = ResourceLimit.loads(resources)
        self.__environ = _process_environ(environ)
        self.__use_sys_env = _process_use_sys_env(use_sys_env)
        _IGlobalConfig.__init__(self, self.__identification, self.__resources, self.__environ, self.__use_sys_env)

    @property
    def identification(self) -> Identification:
        return self.__identification

    @property
    def resources(self) -> ResourceLimit:
        return self.__resources

    @property
    def environ(self) -> Mapping[str, str]:
        return self.__environ

    @property
    def use_sys_env(self) -> Union[set, bool]:
        return self.__use_sys_env

    def __call__(self, environ: Optional[Mapping[str, str]] = None,
                 environ_after: Optional[Mapping[str, str]] = None, **kwargs) -> 'GlobalConfig':
        """
        generate global config
        :param environ: environment variable
        :param environ_after:
        :param kwargs: other arguments
        :return: global config
        """
        _environ = _load_local_environ(self.__use_sys_env)
        _environ = _process_environ(environ, _environ, enable_ext=True)
        _environ = _process_environ(self.__environ, _environ, enable_ext=True)
        _environ = _process_environ(environ_after, _environ, enable_ext=True)

        return GlobalConfig(
            identification=self.__identification,
            resources=self.__resources, environ=_environ,
        )

    @classmethod
    def loads(cls, data) -> 'GlobalConfigTemplate':
        """
        load global config template from data
        :param data: raw data
        :return: global config template
        """
        data = data or {}
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls(**data)
        else:
            raise TypeError('Json or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))


class GlobalConfig(_IGlobalConfig):
    def __init__(self, identification, resources, environ):
        """
        :param identification: identification
        :param resources: resource limits
        :param environ: environment variable
        """
        self.__identification = identification
        self.__resources = resources
        self.__environ = environ
        _IGlobalConfig.__init__(self, self.__identification, self.__resources, self.__environ, None)

    @property
    def identification(self) -> Identification:
        return self.__identification

    @property
    def resources(self) -> ResourceLimit:
        return self.__resources

    @property
    def environ(self) -> Mapping[str, str]:
        return self.__environ

    def __call__(self):
        """
        get global config information
        :return:
        """
        return self.__identification, self.__resources, self.__environ



import os

from pysyslimit import FilePermission, SystemUser, SystemGroup, chmod, chown


def is_absolute_path(path: str) -> bool:
    """
    check if a path is absolute path
    :param path: path
    :return: True if it is an relative path
    """
    return os.path.isabs(path)


def is_relative_path(path: str) -> bool:
    """
    check if a path is relative path
    :param path: path
    :return: True if it is an relative path
    """
    return not os.path.isabs(path)


def is_inner_relative_path(path: str, allow_root: bool = True) -> bool:
    """
    check if a path is relative path
    :param path: path
    :param allow_root: allow path to be root path (.) it self
    :return: True if it is an relative path
    """
    if is_relative_path(path):
        segments = os.path.normpath(path).split(os.sep)
        if not allow_root and segments == ['.']:
            return False
        else:
            return segments[0] != '..'
    else:
        return False


def makedirs(path: str, privilege=None, user=None, group=None):
    """
    Quick make directories with privileges
    :param path: file path
    :param privilege: privileges, not set when none
    :param user: user, not set when none
    :param group: group, not set when none
    """
    path = os.path.normpath(path)
    privilege = FilePermission.loads(privilege) if privilege else None
    user = SystemUser.loads(user) if user else None
    group = SystemGroup.loads(group) if group else None

    if not os.path.exists(path):
        _upper_path, _file = os.path.split(path)
        if _file:
            makedirs(_upper_path or '.', privilege, user, group)
        else:  # this path is the root path, just like '/' or 'path'
            raise FileNotFoundError('Base dir {dir} not found.'.format(dir=repr(_upper_path or _file)))

        os.makedirs(path, exist_ok=True)
        if privilege:
            chmod(path, privilege)
        if user or group:
            chown(path, user, group)
    else:
        if not os.path.isdir(path):
            raise NotADirectoryError('File {file} is not a directory.'.format(file=repr(path)))



import warnings
from typing import Mapping

from hbutils.string import env_template

from .base import _ITask, _check_task_name, ENV_PJI_TASK_NAME
from .task import Task
from ..base import _process_environ
from ..section import SectionCollectionTemplate
from ...control.model import Identification, ResourceLimit


class TaskTemplate(_ITask):
    def __init__(self, name: str, identification=None, resources=None, environ=None, sections=None):
        """
        :param name: name of task
        :param identification: identification
        :param resources: resource limit
        :param environ: environment variables
        :param sections: section templates
        """
        self.__name = name
        self.__identification = Identification.loads(identification)
        self.__resources = ResourceLimit.loads(resources)
        self.__environ = _process_environ(environ)
        self.__sections = SectionCollectionTemplate.loads(sections)

        _ITask.__init__(self, self.__name, self.__identification, self.__resources, self.__environ, self.__sections)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def identification(self) -> Identification:
        return self.__identification

    @property
    def resources(self) -> ResourceLimit:
        return self.__resources

    @property
    def environ(self) -> Mapping[str, str]:
        return self.__environ

    @property
    def sections(self) -> SectionCollectionTemplate:
        return self.__sections

    def __call__(self, scriptdir: str, identification=None, resources=None, environ=None, **kwargs) -> 'Task':
        """
        generate task object
        :param scriptdir: script directory
        :param identification: identification
        :param resources: resource limit
        :param environ: environment variables
        :param kwargs: other arguments
        :return: task object
        """
        environ = dict(_process_environ(self.__environ, environ, enable_ext=True))
        _name = _check_task_name(env_template(self.__name, environ))
        environ[ENV_PJI_TASK_NAME] = _name

        _identification = Identification.merge(Identification.loads(identification), self.__identification)
        _resources = ResourceLimit.merge(ResourceLimit.loads(resources), self.__resources)

        return Task(
            name=_name,
            identification=_identification, resources=_resources, environ=environ,
            sections=self.__sections(
                scriptdir=scriptdir,
                identification=_identification,
                resources=_resources,
                environ=environ, **kwargs
            )
        )

    @classmethod
    def loads(cls, data) -> 'TaskTemplate':
        """
        load task template object from raw data
        :param data: raw data
        :return: task template object
        """
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls(**data)
        elif isinstance(data, tuple):
            _name, _data = data
            if isinstance(_data, dict):
                if 'name' in _data:
                    warnings.warn(RuntimeWarning(
                        'Name found in {name} \'s second value, it will be ignored.'.format(name=repr(_name))))
                _data['name'] = _name
                return cls(**_data)
            else:
                raise TypeError('Second value of task tuple should be a dict but {actual} found.'.format(
                    actual=repr(type(_data).__name__)))
        else:
            raise TypeError('Json or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))



from .base import FileInput, FileInputTemplate
from .collection import FileInputCollectionTemplate, FileInputCollection
from .copy import CopyFileInput, CopyFileInputTemplate
from .general import FileInputType, load_input_template
from .link import LinkFileInput, LinkFileInputTemplate
from .tag import TagFileInput, TagFileInputTemplate



from abc import ABCMeta, abstractmethod
from typing import Optional

from pysyslimit import chown, chmod
from pysyslimit.models.permission.full import FileUserPermission, FilePermission, FileGroupPermission, FileOtherPermission

from ....control.model import Identification


def _load_privilege(privilege=None) -> Optional[FilePermission]:
    """
    load privilege information from data
    :param privilege: raw privilege data
    :return: privilege object or None
    """
    if privilege is not None:
        try:
            _privilege = FilePermission.loads(privilege)
        except (TypeError, ValueError):
            _privilege = FilePermission(
                FileUserPermission.loads(privilege),
                FileGroupPermission.loads('---'),
                FileOtherPermission.loads('---'),
            )
    else:
        _privilege = None

    return _privilege


def _apply_privilege_and_identification(filename: str, privilege=None, identification=None):
    """
    Apply privilege and identification for file
    :param filename: file path
    :param privilege: file privilege
    :param identification: file identification
    """
    if privilege is not None:
        chmod(filename, privilege, recursive=True)
    if identification is not None:
        _ident = Identification.merge(Identification.load_from_file(filename), identification)
        chown(filename, _ident.user, _ident.group, recursive=True)


class FileInputTemplate(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, *args, **kwargs) -> 'FileInput':
        raise NotImplementedError


class FileInput(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, **kwargs):
        raise NotImplementedError



import codecs
import json
import tempfile
from typing import Mapping, Callable, Tuple, List, Any, Optional

from pysyslimit import chown, chmod

from .base import _ISection
from ..info import SectionInfoMapping
from ..input import FileInputCollection
from ..output import FileOutputCollection
from ...base import _process_environ
from ...command import CommandCollection
from ....control.model import Identification, ResourceLimit, RunResult
from ....utils import wrap_empty

_DEFAULT_WORKDIR = '.'
_SECTION_RESULT = Tuple[bool, List[RunResult], Mapping[str, Any]]


class Section(_ISection):
    def __init__(self, name: str,
                 commands_getter: Callable[..., CommandCollection],
                 identification: Identification, resources: ResourceLimit, environ,
                 inputs_getter: Callable[..., FileInputCollection],
                 outputs_getter: Callable[..., FileOutputCollection],
                 infos_getter: Callable[..., SectionInfoMapping],
                 info_dump: Optional[str]):
        """
        :param name: name of section
        :param commands_getter: command collection getter
        :param identification: identification
        :param resources: resource limits
        :param environ: environment variables
        :param inputs_getter: input collection getter
        :param outputs_getter: output collection getter
        :param infos_getter: information collection getter
        """
        self.__name = name
        self.__commands_getter = commands_getter

        self.__identification = identification
        self.__resources = resources
        self.__environ = _process_environ(environ)

        self.__inputs_getter = inputs_getter
        self.__outputs_getter = outputs_getter
        self.__infos_getter = infos_getter
        self.__info_dump = info_dump

        _ISection.__init__(self, self.__name, self.__identification,
                           self.__resources, self.__environ,
                           self.__inputs_getter(workdir=_DEFAULT_WORKDIR),
                           self.__outputs_getter(workdir=_DEFAULT_WORKDIR),
                           self.__infos_getter(workdir=_DEFAULT_WORKDIR),
                           self.__commands_getter(workdir=_DEFAULT_WORKDIR))

    @property
    def name(self) -> str:
        return self.__name

    @property
    def commands_getter(self) -> Callable[..., CommandCollection]:
        return self.__commands_getter

    @property
    def identification(self) -> Identification:
        return self.__identification

    @property
    def resources(self) -> ResourceLimit:
        return self.__resources

    @property
    def environ(self) -> Mapping[str, str]:
        return self.__environ

    @property
    def inputs_getter(self) -> Callable[..., FileInputCollection]:
        return self.__inputs_getter

    @property
    def outputs_getter(self) -> Callable[..., FileOutputCollection]:
        return self.__outputs_getter

    @property
    def infos_getter(self) -> Callable[..., SectionInfoMapping]:
        return self.__infos_getter

    @property
    def info_dump(self) -> Optional[str]:
        return self.__info_dump

    def __call__(self, section_start: Optional[Callable[['Section'], None]] = None,
                 section_complete: Optional[Callable[['Section', _SECTION_RESULT], None]] = None,
                 info_dump_start: Optional[Callable[['Section', Mapping[str, Any]], None]] = None,
                 info_dump_complete: Optional[Callable[['Section', Mapping[str, Any]], None]] = None,
                 **kwargs) -> _SECTION_RESULT:
        """
        run section
        :return: success or not, result list, information
        """
        wrap_empty(section_start)(self)
        with tempfile.TemporaryDirectory() as workdir:
            chmod(workdir, 'r-x------')
            if self.__identification:
                chown(workdir, user=self.__identification.user, group=self.__identification.group)

            self.__inputs_getter(workdir=workdir)(**kwargs)
            _success, _results = self.__commands_getter(workdir=workdir)(**kwargs)
            if _success:
                self.__outputs_getter(workdir=workdir)(**kwargs)
            _info = self.__infos_getter(workdir=workdir)(**kwargs)
            if self.__info_dump:
                wrap_empty(info_dump_start)(self, _info)
                with codecs.open(self.__info_dump, 'w') as cf:
                    json.dump(_info, cf, indent=4, sort_keys=True)
                wrap_empty(info_dump_complete)(self, _info)

            _return = (_success, _results, _info)

        wrap_empty(section_complete)(self, _return)
        return _return



from .base import FileOutputTemplate, FileOutput
from .collection import FileOutputCollectionTemplate, FileOutputCollection
from .copy import CopyFileOutputTemplate, CopyFileOutput
from .general import FileOutputType, load_output_template
from .tag import TagFileOutputTemplate, TagFileOutput



import io
import time

from .encoding import _try_write, _try_read_to_bytes
from ..model import RunResult
from ..model import TimingContent as _AbstractTimingContent
from ..process import interactive_process
from ...utils import eclosing


class TimingStdin(_AbstractTimingContent):
    pass


class TimingStdout(_AbstractTimingContent):
    pass


class TimingStderr(_AbstractTimingContent):
    pass


def timing_run(args, shell: bool = False, stdin=None, stdout=None, stderr=None,
               environ=None, cwd=None, resources=None, identification=None) -> RunResult:
    """
    Create an timing process with stream
    :param args: arguments for execution
    :param shell: use shell to execute args
    :param stdin: stdin stream (none means nothing)
    :param stdout: stdout stream (none means nothing)
    :param stderr: stderr stream (none means nothing)
    :param environ: environment variables
    :param cwd: new work dir
    :param resources: resource limit
    :param identification: user and group for execution
    :return: run result of this time
    """
    stdin_need_close = not stdin
    stdin = stdin or io.BytesIO()

    stdout_need_close = not stdout
    stdout = stdout or io.BytesIO()

    stderr_need_close = not stderr
    stderr = stderr or io.BytesIO()

    with eclosing(stdin, stdin_need_close) as stdin, \
            eclosing(stdout, stdout_need_close) as stdout, \
            eclosing(stderr, stderr_need_close) as stderr:
        with interactive_process(
                args=args, shell=shell,
                environ=environ, cwd=cwd,
                resources=resources, identification=identification,
        ) as ip:
            _stdin = TimingStdin.loads(_try_read_to_bytes(stdin))
            for _time, _line in _stdin.lines:
                _target_time = ip.start_time + _time
                while time.time() < _target_time and not ip.completed:
                    time.sleep(max(min(0.2, _target_time - time.time()), 0.0))

                try:
                    ip.print_stdin(_line)
                except BrokenPipeError:
                    break

            ip.close_stdin()

            _stdout, _stderr = [], []
            for _time, _tag, _line in ip.output_yield:
                if _tag == 'stdout':
                    _stdout.append((_time, _line))
                elif _tag == 'stderr':
                    _stderr.append((_time, _line))
                else:
                    raise ValueError('Unknown output type - {type}.'.format(type=repr(_time)))

            ip.join()
            _try_write(stdout, TimingStdout.loads(_stdout).dumps())
            _try_write(stderr, TimingStderr.loads(_stderr).dumps())

            return ip.result



from multiprocessing import Lock


class ValueProxy:
    def __init__(self, init_value=None):
        self.__value = init_value
        self.__lock = Lock()

    @property
    def value(self):
        with self.__lock:
            return self.__value

    @value.setter
    def value(self, new_value):
        with self.__lock:
            self.__value = new_value



from typing import List, Tuple, Mapping


def _split_environ(string: str) -> Tuple[str, str]:
    """
    split environment string into name and value
    :param string: original string
    :return: name, value
    """
    _name, _value = string.split('=', 2)
    return _name, _value


def _load_environ(envs: List[str]) -> Mapping[str, str]:
    """
    load environment variables
    :param envs: environments
    :return: env dict
    """
    return dict([_split_environ(env) for env in envs])



from .control import common_process, interactive_process, CommonProcess, InteractiveProcess, TimingContent, common_run, \
    timing_run, mutual_run, ProcessResult, RunResult, RunResultStatus, ResourceLimit, Identification
from .entry import DispatchRunner, load_pji_script
from .service import DispatchTemplate, Dispatch, TaskTemplate, Task



def _section_result_to_json(result):
    _name, (_ok, _runs, _info) = result
    return dict(
        name=_name,
        ok=_ok,
        commands=[item.json for item in _runs],
        information=_info,
    )


def result_to_json(success: bool, result):
    """
    from result to json data
    :param success: success or not
    :param result: full result
    :return: full json
    """
    return dict(
        ok=not not success,
        sections=[_section_result_to_json(item) for item in result],
    )



import resource
from typing import Optional, Union

from bitmath import MiB
from hbutils.scale import size_to_bytes, time_to_duration, size_to_bytes_str

from ...utils import allow_none, get_repr_info

_UNLIMITED = -1


def _rmin(x, y):
    if x == _UNLIMITED:
        return y
    elif y == _UNLIMITED:
        return x
    else:
        return min(x, y)


def _rprocess(cur, new):
    _cur_soft, _cur_hard = cur
    if isinstance(new, tuple):
        _new_soft, _new_hard = new
    else:
        _new_soft, _new_hard = new, new

    _hard = _rmin(_cur_hard, _new_hard)
    _soft = _rmin(_new_soft, _hard)

    return _soft, _hard


_memory_process = allow_none(size_to_bytes)
_duration_process = allow_none(time_to_duration)
_number_process = allow_none(lambda x: x)


class ResourceLimit:
    __RESOURCES = {"max_stack", "max_memory", "max_cpu_time", "max_real_time",
                   "max_process_number", "max_output_size", }

    def __init__(
            self,
            max_stack=None,
            max_memory=None,
            max_cpu_time=None,
            max_real_time=None,
            max_process_number=None,
            max_output_size=None
    ):
        """
        :param max_stack: max stack memory (unit: B)
        :param max_memory: max rss memory memory (unit: B)
        :param max_cpu_time: max cpu time (unit: s)
        :param max_real_time: max real time (unit: s)
        :param max_process_number: max process count
        :param max_output_size: max output size (unit: B)
        """
        self.__max_stack = _memory_process(max_stack)
        self.__max_memory = _memory_process(max_memory)
        self.__max_cpu_time = _duration_process(max_cpu_time)
        self.__max_real_time = _duration_process(max_real_time)
        self.__max_process_number = _number_process(max_process_number)
        self.__max_output_size = _memory_process(max_output_size)

    @property
    def max_stack(self):
        """
        :return: max stack size (size: B)
        """
        return self.__max_stack

    @property
    def max_memory(self):
        """
        :return: max rss memory size (size: B)
        """
        return self.__max_memory

    @property
    def max_cpu_time(self):
        """
        :return: max cpu time (unit: s)
        """
        return self.__max_cpu_time

    @property
    def max_real_time(self):
        """
        :return: max real time (unit: s)
        """
        return self.__max_real_time

    @property
    def max_process_number(self):
        """
        :return: max process count
        """
        return self.__max_process_number

    @property
    def max_output_size(self):
        """
        :return: max output size (unit: B)
        """
        return self.__max_output_size

    @classmethod
    def __apply_limit(cls, limit_type, value):
        """
        apply one of the resources
        :param limit_type: type of limitation
        :param value: limitation value
        """
        _limit_value = _rprocess(resource.getrlimit(limit_type), value)
        resource.setrlimit(limit_type, _limit_value)

    def __apply_max_stack(self):
        """
        apply max stack limit
        """
        if self.max_stack:
            real = self.max_stack
        else:
            real = _UNLIMITED
        self.__apply_limit(resource.RLIMIT_STACK, real)

    def __apply_max_memory(self):
        """
        apply max rss memory limit
        """
        if self.max_memory:
            real = round(self.max_memory + MiB(256).bytes)
        else:
            real = _UNLIMITED
        self.__apply_limit(resource.RLIMIT_AS, real)

    def __apply_max_cpu_time(self):
        """
        apply max cpu time limit
        """
        if self.max_cpu_time:
            real = round(self.max_cpu_time) + 1
        else:
            real = _UNLIMITED
        self.__apply_limit(resource.RLIMIT_CPU, real)

    def __apply_max_process_number(self):
        """
        apply max process number limit
        """
        if self.max_process_number:
            real = self.max_process_number
        else:
            real = _UNLIMITED
        self.__apply_limit(resource.RLIMIT_NPROC, real)

    def __apply_max_output_size(self):
        """
        apply max output size limit
        """
        if self.max_output_size:
            real = self.max_output_size
        else:
            real = _UNLIMITED
        self.__apply_limit(resource.RLIMIT_FSIZE, real)

    @property
    def json(self):
        """
        get json format data
        :return: json format data
        """
        return {
            "max_memory": self.max_memory,
            "max_stack": self.max_stack,
            "max_process_number": self.max_process_number,
            "max_output_size": self.max_output_size,
            "max_cpu_time": self.max_cpu_time,
            "max_real_time": self.max_real_time,
        }

    @classmethod
    def load_from_json(cls, json_data: dict) -> 'ResourceLimit':
        """
        load object from json data
        :param json_data: json data
        :return: resource limit object
        """
        return cls(**cls.__filter_by_properties(**json_data))

    @classmethod
    def loads(cls, data: Optional[Union[dict, 'ResourceLimit']]) -> 'ResourceLimit':
        """
        load object from json data or resource limit object
        :param data: json data or object
        :return: resource limit object
        """
        data = data or {}
        if isinstance(data, ResourceLimit):
            return data
        elif isinstance(data, dict):
            return ResourceLimit.load_from_json(json_data=data)
        else:
            raise TypeError('{rl} or {dict} expected, but {actual} found.'.format(
                rl=ResourceLimit.__name__,
                dict=dict.__name__,
                actual=type(data).__name__,
            ))

    def apply(self):
        """
        apply the resource limits
        """
        self.__apply_max_process_number()
        self.__apply_max_stack()
        self.__apply_max_memory()
        self.__apply_max_output_size()
        self.__apply_max_cpu_time()

    @classmethod
    def __filter_by_properties(cls, **kwargs):
        """
        filter the arguments by properties
        :param kwargs: original arguments
        :return: filtered arguments
        """
        return {
            key: value for key, value in kwargs.items() if key in cls.__RESOURCES
        }

    @classmethod
    def merge(cls, *limits: 'ResourceLimit'):
        """
        merge some of the limits
        :param limits: list of the limits
        :return: merged limits
        """

        def _get_min_limitation(array, property_method):
            _values = [_item for _item in [property_method(_item) for _item in array if _item] if _item is not None]
            if _values:
                return min(_values)
            else:
                return None

        _max_stack = _get_min_limitation(limits, lambda _item: _item.max_stack)
        _max_memory = _get_min_limitation(limits, lambda _item: _item.max_memory)
        _max_cpu_time = _get_min_limitation(limits, lambda _item: _item.max_cpu_time)
        _max_real_time = _get_min_limitation(limits, lambda _item: _item.max_real_time)
        _max_process_number = _get_min_limitation(limits, lambda _item: _item.max_process_number)
        _max_output_size = _get_min_limitation(limits, lambda _item: _item.max_output_size)

        return cls(
            max_stack=_max_stack,
            max_memory=_max_memory,
            max_cpu_time=_max_cpu_time,
            max_real_time=_max_real_time,
            max_process_number=_max_process_number,
            max_output_size=_max_output_size,
        )

    def __tuple(self):
        return self.__max_stack, self.__max_memory, self.__max_cpu_time, \
               self.__max_real_time, self.__max_process_number, self.__max_output_size

    def __eq__(self, other):
        if other is self:
            return True
        elif isinstance(other, self.__class__):
            return self.__tuple() == other.__tuple()
        else:
            return False

    def __hash__(self):
        return hash(self.__tuple())

    def __repr__(self):
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('cpu time', (lambda: '%.3fs' % self.max_cpu_time, lambda: self.max_cpu_time is not None)),
                ('real time', (lambda: '%.3fs' % self.max_real_time, lambda: self.max_real_time is not None)),
                ('memory', (lambda: size_to_bytes_str(self.max_memory), lambda: self.max_memory is not None)),
                ('stack', (lambda: size_to_bytes_str(self.max_stack), lambda: self.max_stack is not None)),
                ('process', (lambda: self.max_process_number, lambda: self.max_process_number is not None)),
                ('output size',
                 (lambda: size_to_bytes_str(self.max_output_size), lambda: self.max_output_size is not None)),
            ]
        )



import os
from typing import Union, List, Optional

from hbutils.string import env_template

from .base import _ICommandBase, CommandMode
from .command import Command
from ..base import _check_workdir_position, _process_environ
from ...control.model import ResourceLimit, Identification


class CommandTemplate(_ICommandBase):
    __DEFAULT_WORKDIR = '.'
    __DEFAULT_MODE = CommandMode.COMMON

    def __init__(self, args: Union[str, List[str]], shell: bool = True,
                 workdir: Optional[str] = None, resources=None,
                 mode=None, stdin=None, stdout=None, stderr=None):
        """
        :param args: arguments
        :param shell: use shell mode
        :param workdir: work directory
        :param resources: resource limits
        :param mode: command mode value
        :param stdin: stdin file
        :param stdout: stdout file
        :param stderr: stderr file
        """
        if not isinstance(args, (str, list)):
            raise TypeError('Args should be str or list but {actual} found.'.format(actual=repr(type(args).__name__)))
        if shell and not isinstance(args, str):
            raise ValueError(
                'Args should be string when shell is true but {actual} found.'.format(actual=repr(type(args).__name__)))
        self.__args = args
        self.__shell = shell

        self.__workdir = str(workdir or self.__DEFAULT_WORKDIR)
        self.__resources = ResourceLimit.loads(resources or {})

        self.__mode = CommandMode.loads(mode or self.__DEFAULT_MODE)
        self.__stdin = stdin
        self.__stdout = stdout
        self.__stderr = stderr

        _ICommandBase.__init__(self, self.__args, self.__shell, self.__workdir,
                               None, self.__resources, self.__mode)

    @property
    def args(self) -> Union[str, List[str]]:
        return self.__args

    @property
    def shell(self) -> bool:
        return self.__shell

    @property
    def workdir(self) -> str:
        return self.__workdir

    @property
    def resources(self) -> ResourceLimit:
        return self.__resources

    @property
    def mode(self) -> CommandMode:
        return self.__mode

    @property
    def stdin(self):
        return self.__stdin

    @property
    def stdout(self):
        return self.__stdout

    @property
    def stderr(self):
        return self.__stderr

    def __tuple(self):
        return self.__args, self.__shell, self.__workdir, self.__resources, \
               self.__mode, self.__stdin, self.__stdout, self.__stdout

    def __eq__(self, other):
        """
        check equality
        :param other: another command template object
        :return:
        """
        if other is self:
            return True
        elif isinstance(other, self.__class__):
            return self.__tuple() == other.__tuple()
        else:
            return False

    def __hash__(self):
        """
        get hash value of object
        :return: hash value
        """
        return hash(self.__tuple())

    def __call__(self, identification=None, resources=None, workdir=None, environ=None, **kwargs) -> Command:
        """
        get command object from template
        :param identification: identification
        :param resources: resource limits
        :param workdir: work directory
        :param environ: environment variables
        :return: command object
        """
        environ = _process_environ(environ)
        _identification = Identification.loads(identification or {})
        _resources = ResourceLimit.merge(ResourceLimit.loads(resources or {}), self.__resources)
        _workdir = os.path.normpath(
            os.path.join(workdir or '.', _check_workdir_position(env_template(self.__workdir, environ))))
        _stdin = os.path.normpath(os.path.join(_workdir, env_template(self.__stdin, environ))) \
            if isinstance(self.__stdin, str) else self.__stdin
        _stdout = os.path.normpath(os.path.join(_workdir, env_template(self.__stdout, environ))) \
            if isinstance(self.__stdout, str) else self.__stdout
        _stderr = os.path.normpath(os.path.join(_workdir, env_template(self.__stderr, environ))) \
            if isinstance(self.__stderr, str) else self.__stderr

        return Command(
            args=self.__args, shell=self.__shell,
            workdir=_workdir, environ=environ,
            identification=_identification, resources=_resources,
            mode=self.__mode, stdin=_stdin, stdout=_stdout, stderr=_stderr,
        )

    @classmethod
    def loads(cls, data) -> 'CommandTemplate':
        """
        load command template from data
        :param data: raw data
        :return: command template object
        """
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls(**data)
        elif isinstance(data, str):
            return cls(args=data)
        elif isinstance(data, (list, tuple)):
            return cls(args=list(data), shell=False)
        else:
            raise TypeError('Json or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))



import collections
from typing import Iterable, TypeVar, Set

_T = TypeVar('_T')


def duplicates(array: Iterable[_T]) -> Set[_T]:
    return {item for item, count in collections.Counter(array).items() if count > 1}



from typing import List, Tuple


def get_repr_info(cls: type, args: List[Tuple]):
    """
    get representation information
    :param cls: class object
    :param args: arguments to display
    :return: representation string
    """
    _data_items = []
    for item in args:
        if isinstance(item, tuple):
            if len(item) == 2:
                name, fd = item
                if isinstance(fd, tuple):
                    _data_func, _present_func = fd
                else:
                    _data_func, _present_func = fd, lambda: True
            elif len(item) == 3:
                name, _data_func, _present_func = item
            else:
                raise ValueError('Tuple\'s length should be 2 or 3 but {actual} found.'.format(actual=repr(len(item))))

            if _present_func():
                _data_items.append('{name}: {data}'.format(name=name, data=_data_func()))
        else:
            raise TypeError(
                'Argument item should be tuple but {actual} found.'.format(actual=repr(type(item).__name__)))

    if _data_items:
        return '<{cls} {data}>'.format(cls=cls.__name__, data=', '.join(_data_items))
    else:
        return '<{cls}>'.format(cls=cls.__name__)



import os
import pickle
import sys
import time
from multiprocessing import Value
from multiprocessing.synchronize import Event as EventClass
from typing import Mapping

import where

from ...utils import args_split


class ExecutorException(Exception):
    def __init__(self, exception):
        self.__exception = exception
        Exception.__init__(self, repr(exception))

    @property
    def exception(self):
        return self.__exception


def get_child_executor_func(args, environ: Mapping[str, str], preexec_fn,
                            executor_prepare_ok: EventClass, exception_pipes,
                            parent_initialized: EventClass,
                            start_time_ok: EventClass, start_time: Value,
                            stdin_pipes, stdout_pipes, stderr_pipes):
    args = args_split(args)
    arg_file = where.first(args[0])

    if not arg_file:
        raise EnvironmentError('Executable {exec} not found.'.format(exec=args[0]))

    stdin_read, stdin_write = stdin_pipes
    stdout_read, stdout_write = stdout_pipes
    stderr_read, stderr_write = stderr_pipes
    exception_read, exception_write = exception_pipes

    # noinspection DuplicatedCode
    def _execute_child():
        os.close(stdin_write)
        sys.stdin = sys.__stdin__
        os.dup2(stdin_read, sys.stdin.fileno())

        os.close(stdout_read)
        sys.stdout = sys.__stdout__
        os.dup2(stdout_write, sys.stdout.fileno())

        os.close(stderr_read)
        sys.stderr = sys.__stderr__
        os.dup2(stderr_write, sys.stderr.fileno())

        _exception = None
        try:
            if preexec_fn is not None:
                preexec_fn()
        except Exception as err:
            _exception = err

        os.close(exception_read)
        with os.fdopen(exception_write, 'wb', 0) as ef:
            if _exception is not None:
                pickle.dump(ExecutorException(_exception), ef)
            else:
                pickle.dump(None, ef)
        executor_prepare_ok.set()

        if _exception is None:
            parent_initialized.wait()
            start_time.value = time.time()
            start_time_ok.set()

            os.execvpe(arg_file, args, environ)

    return _execute_child



import importlib
import io
import os
import sys
import time
from multiprocessing import Process, Event
from threading import Thread
from typing import Callable

from .encoding import _auto_encode, _try_write
from ..model import TimingContent, RunResult
from ..process import interactive_process
from ...utils import eclosing


class MutualStdout(TimingContent):
    pass


class MutualStderr(TimingContent):
    pass


_INTERACT_FUNC = Callable[[], None]


def _load_func_from_str(source: str) -> _INTERACT_FUNC:
    _split = source.split(':')
    if len(_split) == 2:
        _module_name, _func_name = _split
        _module = importlib.import_module(_module_name)
        _func = getattr(_module, _func_name)
        if hasattr(_func, '__call__'):
            return _func
        else:
            raise TypeError('Callable expected but {actual} found.'.format(actual=repr(type(_func).__name__)))
    else:
        raise ValueError('Invalid mutual function - {func}.'.format(func=repr(source)))


def _load_func(func) -> _INTERACT_FUNC:
    if func is None:
        return lambda: None
    elif hasattr(func, '__call__'):
        return func
    elif isinstance(func, str):
        return _load_func_from_str(func)
    else:
        raise TypeError('Callable or str source expected but {actual} found.'.format(actual=repr(type(func).__name__)))


def mutual_run(args, shell: bool = False, stdin=None, stdout=None, stderr=None,
               environ=None, cwd=None, resources=None, identification=None) -> RunResult:
    """
    Create an mutual process with stream
    :param args: arguments for execution
    :param shell: use shell to execute args
    :param stdin: stdin interaction function (none means nothing)
    :param stdout: stdout stream (none means nothing)
    :param stderr: stderr stream (none means nothing)
    :param environ: environment variables
    :param cwd: new work dir
    :param resources: resource limit
    :param identification: user and group for execution
    :return: run result of this time
    """
    stdin = _load_func(stdin)

    stdout_need_close = not stdout
    stdout = stdout or io.BytesIO()

    stderr_need_close = not stderr
    stderr = stderr or io.BytesIO()

    with eclosing(stdout, stdout_need_close) as stdout, \
            eclosing(stderr, stderr_need_close) as stderr:
        mutual_stdout_get, mutual_stdout_put = os.pipe()
        mutual_stdin_get, mutual_stdin_put = os.pipe()
        mutual_stderr_get, mutual_stderr_put = os.pipe()

        _mutual_initialize_ok = Event()
        _prepare_ok = Event()

        # noinspection DuplicatedCode
        def _launch_mutual():
            os.close(mutual_stdin_put)
            # sys.stdin = sys.__stdin__  # this stdin is closed, ValueError: I/O operation on closed file
            os.dup2(mutual_stdin_get, sys.stdin.fileno())

            os.close(mutual_stdout_get)
            sys.stdout = sys.__stdout__
            os.dup2(mutual_stdout_put, sys.stdout.fileno())

            os.close(mutual_stderr_get)
            sys.stderr = sys.__stderr__
            os.dup2(mutual_stderr_put, sys.stderr.fileno())

            _mutual_initialize_ok.set()
            _prepare_ok.wait()
            stdin()

        _mutual_process = Process(target=_launch_mutual)

        with interactive_process(
                args=args, shell=shell,
                environ=environ, cwd=cwd,
                resources=resources, identification=identification,
        ) as ip:
            _mutual_process.start()
            os.close(mutual_stdin_get)
            os.close(mutual_stdout_put)
            os.close(mutual_stderr_put)
            _mutual_initialize_ok.wait()

            _stdout_list = []
            _stderr_list = []

            with os.fdopen(mutual_stdin_put, 'wb', 0) as f_mutual_stdin, \
                    os.fdopen(mutual_stdout_get, 'rb', 0) as f_mutual_stdout, \
                    os.fdopen(mutual_stderr_get, 'rb', 0) as f_mutual_stderr:
                def _load_output_from_ip():
                    _mutual_close_stdin = False
                    for _time, _tag, _line in ip.output_yield:
                        if _tag == 'stdout':
                            if not _mutual_close_stdin:
                                _mutual_initialize_ok.wait()
                                try:
                                    f_mutual_stdin.write(_line + _auto_encode(os.linesep))
                                    f_mutual_stdin.flush()
                                except BrokenPipeError:
                                    _mutual_close_stdin = True
                        elif _tag == 'stderr':
                            _stderr_list.append((time.time() - ip.start_time, _line))

                    f_mutual_stdin.close()

                def _load_stdout_from_mutual():
                    _mutual_initialize_ok.wait()
                    for _line in f_mutual_stdout:
                        try:
                            ip.print_stdin(_line.rstrip(b'\r\n'))
                        except BrokenPipeError:
                            break
                    ip.close_stdin()

                def _load_stderr_from_mutual():
                    _mutual_initialize_ok.wait()
                    for _line in f_mutual_stderr:
                        _stdout_list.append((time.time() - ip.start_time, _line.rstrip(b'\r\n')))

                _threads = [Thread(target=_func) for _func in
                            [_load_output_from_ip, _load_stdout_from_mutual, _load_stderr_from_mutual]]
                for _item in _threads:
                    _item.start()
                _prepare_ok.set()

                for _item in _threads:
                    _item.join()

            _try_write(stdout, MutualStdout(_stdout_list).dumps())
            _try_write(stderr, MutualStderr(_stderr_list).dumps())

            return ip.result



import os
import signal
import time
from abc import ABCMeta
from multiprocessing import Event, Value, Lock
from multiprocessing.synchronize import Event as EventClass
from queue import Empty, Queue
from threading import Thread
from typing import Tuple, Callable, Optional

from ..model import ProcessResult, ResourceLimit, RunResult, RunResultStatus
from ...utils import ValueProxy


class GeneralProcess(metaclass=ABCMeta):
    def __init__(self, start_time: float, resources: ResourceLimit,
                 process_result_func: Callable[[], Optional[ProcessResult]],
                 lifetime_event: EventClass, lock: Optional[Lock] = None):
        self.__start_time = start_time
        self.__resources = resources
        self.__process_result = None
        self.__process_result_func = process_result_func
        self.__result = None
        self.__lifetime_event = lifetime_event
        self.__lock = lock or Lock()

    def __get_process_result(self) -> ProcessResult:
        if self.__process_result is None:
            self.__process_result = self.__process_result_func()
        return self.__process_result

    def __get_result(self) -> RunResult:
        if self.__result is None or self.__result.result is None:
            self.__result = RunResult(self.__resources, self.__get_process_result())
        return self.__result

    def _wait_for_end(self):
        self.__lifetime_event.wait()

    @property
    def start_time(self) -> float:
        with self.__lock:
            return self.__start_time

    @property
    def resources(self) -> ResourceLimit:
        with self.__lock:
            return self.__resources

    @property
    def result(self) -> RunResult:
        with self.__lock:
            return self.__get_result()

    @property
    def ok(self) -> bool:
        with self.__lock:
            return self.__get_result().ok

    @property
    def completed(self) -> bool:
        with self.__lock:
            return self.__get_result().completed

    @property
    def status(self) -> RunResultStatus:
        with self.__lock:
            return self.__get_result().status

    def join(self):
        with self.__lock:
            self._wait_for_end()


BYTES_LINESEQ = bytes(os.linesep, 'utf8')


def read_all_from_bytes_stream(stream) -> bytes:
    return b''.join([line for line in stream])


def load_lines_from_bytes_stream(stream, loader_initialized: EventClass, queue: Queue, transformer=None):
    _middle_queue = Queue()
    _output_load_complete = Event()
    transformer = transformer or (lambda x: x)

    def _output_load_func():
        loader_initialized.set()
        for line in stream:
            _middle_queue.put((time.time(), line))

        _output_load_complete.set()

    def _item_process_func():
        while not _middle_queue.empty() or not _output_load_complete.is_set():
            try:
                item = _middle_queue.get(timeout=0.2)
            except Empty:
                continue
            else:
                queue.put(transformer(item))

    _output_load_thread = Thread(target=_output_load_func)
    _item_process_func = Thread(target=_item_process_func)

    _output_load_thread.start()
    _item_process_func.start()

    _output_load_thread.join()
    _item_process_func.join()


def measure_thread(start_time_ok: Event, start_time: Value, child_pid: int) \
        -> Tuple[Thread, EventClass, EventClass, EventClass, ValueProxy]:
    _process_result = ValueProxy()
    _process_complete = Event()
    _measure_initialized = Event()
    _measure_complete = Event()

    def _thread_func():
        _measure_initialized.set()
        _, status, resource_usage = os.wait4(child_pid, os.WSTOPPED)
        _process_complete.set()
        start_time_ok.wait()
        _process_result.value = ProcessResult(status, start_time.value, time.time(), resource_usage)
        _measure_complete.set()

    return Thread(target=_thread_func), _measure_initialized, _process_complete, _measure_complete, _process_result


def killer_thread(start_time_ok: Event, start_time: Value, child_pid: int,
                  real_time_limit: float, process_complete: Event) -> Tuple[Thread, EventClass]:
    _killer_initialized = Event()

    def _thread_func():
        _killer_initialized.set()
        if real_time_limit is not None:
            start_time_ok.wait()
            target_time = start_time.value + real_time_limit
            while time.time() < target_time and not process_complete.is_set():
                time.sleep(min(max(target_time - time.time(), 0.0), 0.2))
            if not process_complete.is_set():
                os.kill(child_pid, signal.SIGKILL)
                process_complete.wait()

    return Thread(target=_thread_func), _killer_initialized



import io
import json
import math
import os
import re
from abc import ABCMeta
from typing import List, Tuple, Optional, Union, TypeVar, Type

from ...utils import auto_encode_support, auto_decode_support, auto_load_json, JsonLoadError, get_repr_info

_auto_encode = auto_encode_support(lambda x: x)
_auto_decode = auto_decode_support(lambda x: x)

_LINE_TYPING = Union[bytes, bytearray, str]
_TIMING_LINE_TYPING = Tuple[float, _LINE_TYPING]
_TIMING_LIST_TYPING = List[_TIMING_LINE_TYPING]


def _stable_process(lines: _TIMING_LIST_TYPING) -> _TIMING_LIST_TYPING:
    lines = [(float(_time), _auto_encode(_line).rstrip(b'\r\n')) for _time, _line in lines]
    _sorted = sorted([(_time, i, _line) for i, (_time, _line) in enumerate(lines)])
    return [(_time, _line) for _time, i, _line in _sorted]


class TimingContent(metaclass=ABCMeta):
    def __init__(self, lines: Optional[_TIMING_LIST_TYPING] = None):
        self.__lines = _stable_process(lines or [])

    @property
    def lines(self) -> List[Tuple[float, bytes]]:
        return list(self.__lines)

    @property
    def str_lines(self) -> List[Tuple[float, str]]:
        return [(_time, _auto_decode(_line)) for _time, _line in self.__lines]

    @classmethod
    def load(cls, stream) -> 'TimingContent':
        return _load_from_stream(stream, cls)

    @classmethod
    def loads(cls, data) -> 'TimingContent':
        return _load_from_data(data, cls)

    def to_json(self):
        return [{
            'time': _time,
            'line': _auto_decode(_line),
        } for _time, _line in self.__lines]

    __DUMP_FLOAT_LENGTH = 6

    def dumps(self) -> str:
        if self.__lines:
            _end_time = self.__lines[-1][0]
            _int_length = int(math.floor(math.log10(_end_time)) + 1)
            _mask = '%%.%sf' % self.__DUMP_FLOAT_LENGTH
            with io.StringIO() as s:
                for _time, _line in self.__lines:
                    _time_str = _mask % _time
                    _time_str = ' ' * (self.__DUMP_FLOAT_LENGTH + 1 + _int_length - len(_time_str)) + _time_str
                    s.write('[%s]%s%s' % (_time_str, _auto_decode(_line), _auto_decode(os.linesep)))

                return s.getvalue()
        else:
            return ''

    def dump(self, stream):
        _str_content = self.dumps()
        _bytes_content = _auto_encode(_str_content)
        try:
            stream.write(_bytes_content)
        except TypeError:
            stream.write(_str_content)

    def __eq__(self, other):
        if other is self:
            return True
        elif isinstance(other, TimingContent):
            return self.__lines == other.__lines
        else:
            return False

    def __repr__(self):
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('lines', lambda: len(self.__lines)),
                ('start_time', (lambda: '%.3fs' % self.__lines[0][0], lambda: len(self.__lines) > 0)),
                ('end_time', (lambda: '%.3fs' % self.__lines[-1][0], lambda: len(self.__lines) > 0)),
            ]
        )


_LINE_IDENT = (re.compile(r'\s*\[\s*(\d+(\.\d*)?)\s*]([\s\S]*)'), (1, 3))
_LINE_COMMENT = re.compile(r'\s*#\s*[\s\S]*')

_TS = TypeVar('_TS', bound=TimingContent)


def _load_from_line_ident(stream, cls: Type[_TS]) -> _TS:
    _result = []
    for line in stream:
        _str_line = _auto_decode(line).rstrip('\r\n')
        if _str_line.strip() and not _LINE_COMMENT.fullmatch(_str_line):
            _pattern, (_gtime, _gline) = _LINE_IDENT
            _match = _pattern.fullmatch(_str_line)
            if _match:
                _time, _line = float(_match[_gtime]), _auto_encode(_match[_gline])
                _result.append((_time, _line))
            else:
                raise ValueError('Invalid line {line} for timing script.'.format(line=repr(_str_line)))

    return cls(_result)


def _load_from_json(stream, cls: Type[_TS]) -> _TS:
    _json = auto_load_json(stream)
    if not isinstance(_json, list):
        raise TypeError('Timing script json should be a list but {actual} found.'.format(actual=type(_json).__name__))

    _result = []
    for _item in _json:
        if not isinstance(_item, dict):
            raise TypeError('Timing line should be a dict but {actual} found.'.format(actual=type(_item).__name__))

        _time = _item['time']
        try:
            _lines = _item['line']
        except KeyError:
            _lines = _item['lines']

        if isinstance(_lines, list):
            _lines = list(_lines)
        elif isinstance(_lines, dict):
            raise TypeError('Line should not be a dict.')
        else:
            _lines = [str(_lines)]

        for _line in _lines:
            _result.append((_time, _auto_encode(_line)))

    return cls(_result)


_METHODS = [
    (_load_from_line_ident, ValueError),
    (_load_from_json, (JsonLoadError, TypeError, KeyError))
]


def _load_from_stream(stream, cls: Type[_TS]) -> _TS:
    _init_position = stream.tell()
    _last_err = None

    for _func, _except in _METHODS:
        try:
            stream.seek(_init_position)
            return _func(stream, cls)
        except _except as err:
            _last_err = err

    raise _last_err


def _load_from_data(data, cls: Type[_TS]) -> _TS:
    if isinstance(data, cls):
        return data
    elif isinstance(data, list) or data is None:
        data = data or []
        if data and isinstance(data[0], tuple):
            return cls(data)
        else:
            with io.BytesIO(_auto_encode(json.dumps(data))) as b:
                return _load_from_json(b, cls)
    elif isinstance(data, str):
        with io.StringIO(data) as file:
            return cls.load(file)
    elif isinstance(data, (bytes, bytearray)):
        with io.BytesIO(bytes(data)) as file:
            return cls.load(file)
    else:
        raise TypeError('{cls}, list, str or bytes expected but {actual} found.'.format(
            cls=cls.__name__,
            actual=repr(type(data).__name__),
        ))



from typing import Optional, Mapping, Type

from .event import _load_dispatch_getter
from ..runner import DispatchRunner


def load_pji_script(script: str, runner_class: Optional[Type[DispatchRunner]] = None):
    """
    load pji script from script file
    :param script: script file or dir
    :param runner_class: runner class for executing
    :return: script getter (pass in environ and environ_after before use it)
    """
    _dispatch_getter = _load_dispatch_getter(script, runner_class)

    def _run_task(task_name: str,
                  environ: Optional[Mapping[str, str]] = None,
                  environ_after: Optional[Mapping[str, str]] = None):
        return _dispatch_getter(
            environ=environ,
            environ_after=environ_after,
        )(task_name)

    return _run_task



import codecs
import os
from functools import partial
from typing import Type, Callable

from .runner import DispatchRunner
from ..service import DispatchTemplate, Dispatch
from ..utils import auto_load_json

_DEFAULT_FILENAME = 'pscript.yml'


def _load_dispatch_template(filename: str) -> DispatchTemplate:
    with codecs.open(filename, 'r') as file:
        _json = auto_load_json(file)

    return DispatchTemplate.loads(_json)


def _load_dispatch_getter(runner_class: Type[DispatchRunner], filename: str = None) \
        -> Callable[..., Dispatch]:
    filename = filename or _DEFAULT_FILENAME
    if os.path.isdir(filename):
        filename = os.path.join(filename, _DEFAULT_FILENAME)

    _dir, _ = os.path.split(os.path.normpath(os.path.abspath(filename)))

    return partial(runner_class(_load_dispatch_template(filename)), scriptdir=_dir)



import re
from abc import ABCMeta

from hbutils.string import truncate

from ...control.model import Identification, ResourceLimit
from ...utils import get_repr_info

_TASK_NAME_PATTERN = re.compile('[a-zA-Z_][a-zA-Z0-9_]*')


def _check_task_name(name: str) -> str:
    """
    check task name valid or not
    :param name: task name
    :return: task name
    """
    if _TASK_NAME_PATTERN.fullmatch(name):
        return name
    else:
        raise ValueError('Task name should match {pattern} but {actual} found.'.format(
            pattern=repr(_TASK_NAME_PATTERN.pattern),
            actual=repr(name),
        ))


ENV_PJI_TASK_NAME = 'PJI_TASK_NAME'


class _ITask(metaclass=ABCMeta):
    def __init__(self, name: str, identification=None, resources=None, environ=None, sections=None):
        """
        :param name: name of task
        :param identification: identification
        :param resources: resource limit
        :param environ: environment variables
        :param sections: sections
        """
        self.__name = name
        self.__identification = identification
        self.__resources = resources
        self.__environ = environ
        self.__sections = sections

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('name', lambda: repr(self.__name)),
                ('identification',
                 lambda: truncate(repr(self.__identification), width=48, show_length=True, tail_length=16),
                 lambda: self.__identification and self.__identification != Identification.loads({})),
                ('resources', lambda: truncate(repr(self.__resources), width=64, show_length=True, tail_length=16),
                 lambda: self.__resources != ResourceLimit.loads({})),
                ('sections', lambda: truncate(repr(self.__sections), width=64, show_length=True, tail_length=16),
                 lambda: self.__sections),
            ]
        )



import os
import pickle
from multiprocessing import Event, Value, Lock
from multiprocessing.synchronize import Event as EventClass
from queue import Queue, Empty
from threading import Thread
from typing import Optional, Mapping

from .base import BYTES_LINESEQ, measure_thread, killer_thread, load_lines_from_bytes_stream, GeneralProcess
from .decorator import process_setter
from .executor import get_child_executor_func
from ..model import ResourceLimit
from ...utils import gen_lock


class InteractiveProcess(GeneralProcess):
    def __init__(self, start_time: float, stdin_stream, output_iter,
                 resources: ResourceLimit, process_result_func, lifetime_event: EventClass):
        self.__lock = Lock()
        GeneralProcess.__init__(self, start_time, resources, process_result_func, lifetime_event, self.__lock)

        self.__stdin_stream = stdin_stream
        self.__output_iter = output_iter
        self.__stdin_closed = False

    def __write_stdin(self, data: bytes):
        try:
            self.__stdin_stream.write(data)
        except BrokenPipeError as err:
            self.__stdin_closed = True
            raise err

    def __flush_stdin(self):
        try:
            self.__stdin_stream.flush()
        except BrokenPipeError as err:
            self.__stdin_closed = True
            raise err

    def __close_stdin(self):
        try:
            self.__stdin_stream.close()
            self.__stdin_closed = True
        except BrokenPipeError as err:
            self.__stdin_closed = True
            raise err

    def __load_all_output(self):
        _ = list(self.__output_iter)

    def __exit(self):
        if not self.__stdin_closed:
            self.__close_stdin()
        self.__load_all_output()
        self._wait_for_end()

    @property
    def output_yield(self):
        with self.__lock:
            return self.__output_iter

    def print_stdin(self, line: bytes, flush: bool = True, end: bytes = BYTES_LINESEQ):
        with self.__lock:
            self.__write_stdin(line + end)
            if flush:
                self.__flush_stdin()

    def close_stdin(self):
        with self.__lock:
            self.__close_stdin()

    def __enter__(self):
        with self.__lock:
            return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.__lock:
            self.__exit()


def _read_pipe(pipe_entry, start_time_ok: EventClass, start_time: Value,
               tag: str, loader_initialized: EventClass, queue: Queue):
    def _transform_func(item):
        _time, _line = item
        start_time_ok.wait()
        return _time - start_time.value, tag, _line.rstrip(b'\r\n')

    with os.fdopen(pipe_entry, 'rb', 0) as stream:
        return load_lines_from_bytes_stream(stream, loader_initialized, queue, _transform_func)


# Attention: only real_time_limit will be processed in this function, other limits will be processed in decorator
# noinspection DuplicatedCode, PyIncorrectDocstring,PyUnresolvedReferences,SpellCheckingInspection
@process_setter
def interactive_process(args, preexec_fn=None, resources=None,
                        environ: Optional[Mapping[str, str]] = None) -> InteractiveProcess:
    """
    Create an interactive process
    :param args: arguments for execution
    :param shell: use shell to execute args
    :param preexec_fn: pre execute function to attach before that
    :param resources: resource limit
    :param environ: environment variables
    :param cwd: new work dir
    :param identification: user and group for execution
    :return: InteractiveProcess object to do run
    """
    resources = ResourceLimit.loads(resources)
    _full_lifetime_complete = Event()
    environ = dict(environ or {})

    _executor_prepare_ok = Event()
    _exception_get, _exception_put = os.pipe()

    _parent_initialized = Event()
    _start_time = Value('d', 0.0)
    _start_time_ok = Event()

    # noinspection DuplicatedCode
    def _execute_parent() -> InteractiveProcess:
        os.close(stdin_read)
        os.close(stdout_write)
        os.close(stderr_write)

        # measure thread
        _measure_thread, _measure_initialized, _process_complete, _measure_complete, _result_proxy = measure_thread(
            start_time_ok=_start_time_ok,
            start_time=_start_time,
            child_pid=child_pid,
        )

        # killer thread
        _killer_thread, _killer_initialized = killer_thread(
            start_time_ok=_start_time_ok,
            start_time=_start_time,
            child_pid=child_pid,
            real_time_limit=resources.max_real_time,
            process_complete=_process_complete,
        )

        # lines output
        _output_queue = Queue()
        _output_start, _output_complete = Event(), Event()
        _stdout_initialized, _stderr_initialized = Event(), Event()
        _stdout_thread = Thread(
            target=lambda: _read_pipe(
                pipe_entry=stdout_read,
                start_time_ok=_start_time_ok,
                start_time=_start_time,
                tag='stdout',
                loader_initialized=_stdout_initialized,
                queue=_output_queue,
            ))
        _stderr_thread = Thread(
            target=lambda: _read_pipe(
                pipe_entry=stderr_read,
                start_time_ok=_start_time_ok,
                start_time=_start_time,
                tag='stderr',
                loader_initialized=_stderr_initialized,
                queue=_output_queue,
            ))

        def _output_queue_func():
            _stdout_thread.start()
            _stderr_thread.start()
            _stdout_initialized.wait()
            _stderr_initialized.wait()
            _output_start.set()

            _stdout_thread.join()
            _stderr_thread.join()
            _output_complete.set()

        _queue_thread = Thread(target=_output_queue_func)

        def _output_yield():
            _output_start.wait()
            while not _output_complete.is_set() or not _output_queue.empty():
                try:
                    _time, _tag, _line = _output_queue.get(timeout=0.2)
                except Empty:
                    pass
                else:
                    yield _time, _tag, _line

            _measure_thread.join()
            _killer_thread.join()
            _queue_thread.join()
            _full_lifetime_complete.set()

        # waiting for prepare ok
        _executor_prepare_ok.wait()
        with os.fdopen(_exception_get, 'rb', 0) as ef:
            _exception = pickle.load(ef)
            if _exception:
                raise _exception

        # start all the threads and services
        _measure_thread.start()
        _killer_thread.start()
        _queue_thread.start()
        _stdin_stream = os.fdopen(stdin_write, 'wb', 0)
        _output_iter = gen_lock(_output_yield())

        # wait for all the thread initialized
        _measure_initialized.wait()
        _killer_initialized.wait()
        _output_start.wait()
        _parent_initialized.set()

        _start_time_ok.wait()
        return InteractiveProcess(
            start_time=_start_time.value,
            stdin_stream=_stdin_stream,
            output_iter=_output_iter,
            resources=resources,
            process_result_func=lambda: _result_proxy.value,
            lifetime_event=_full_lifetime_complete,
        )

    stdin_read, stdin_write = os.pipe()
    stdout_read, stdout_write = os.pipe()
    stderr_read, stderr_write = os.pipe()

    _execute_child = get_child_executor_func(
        args, dict(environ or {}), preexec_fn,
        _executor_prepare_ok, (_exception_get, _exception_put),
        _parent_initialized,
        _start_time_ok, _start_time,
        (stdin_read, stdin_write),
        (stdout_read, stdout_write),
        (stderr_read, stderr_write),
    )

    child_pid = os.fork()

    if not child_pid:
        _execute_child()
    else:
        return _execute_parent()



from abc import ABCMeta
from typing import Mapping

from .task import Task
from .template import TaskTemplate
from ...utils import get_repr_info, duplicates


class _ITaskMapping(metaclass=ABCMeta):
    def __init__(self, tasks):
        """
        :param tasks: tasks ot task templates
        """
        self.__tasks = tasks

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('tasks', lambda: repr(tuple(self.__tasks.keys())))
            ]
        )


class TaskMappingTemplate(_ITaskMapping):
    def __init__(self, *tasks):
        _tasks = [TaskTemplate.loads(_item) for _item in (tasks or [])]
        _duplicated_names = duplicates([_item.name for _item in _tasks])
        if _duplicated_names:
            raise KeyError(
                'Duplicated name for task mapping - {names}'.format(names=repr(sorted(list(_duplicated_names)))))

        self.__tasks = {_task.name: _task for _task in _tasks}
        _ITaskMapping.__init__(self, self.__tasks)

    @property
    def items(self) -> Mapping[str, TaskTemplate]:
        return self.__tasks

    def __iter__(self):
        return self.items.__iter__()

    def __call__(self, scriptdir: str, identification=None, resources=None, environ=None, **kwargs) -> 'TaskMapping':
        _tasks = [_task(
            scriptdir=scriptdir,
            identification=identification,
            resources=resources,
            environ=environ,
        ) for _task in self.__tasks.values()]
        _duplicated_names = duplicates([_item.name for _item in _tasks])
        if _duplicated_names:
            raise KeyError(
                'Duplicated name for task mapping - {names}'.format(names=repr(sorted(list(_duplicated_names)))))

        return TaskMapping(*_tasks)

    @classmethod
    def loads(cls, data) -> 'TaskMappingTemplate':
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls(*[TaskTemplate.loads((_name, _value)) for _name, _value in data.items()])
        elif isinstance(data, (list, tuple)):
            return cls(*data)
        elif isinstance(data, TaskTemplate):
            return cls(data)
        else:
            raise TypeError('Json, array or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))


class TaskMapping(_ITaskMapping):
    def __init__(self, *tasks: Task):
        self.__tasks = {_task.name: _task for _task in tasks}
        _ITaskMapping.__init__(self, self.__tasks)

    @property
    def items(self) -> Mapping[str, Task]:
        return self.__tasks

    def __iter__(self):
        return self.items.__iter__()

    def __call__(self, task_name: str, **kwargs):
        if task_name not in self.__tasks.keys():
            raise KeyError('Task {task} not found.'.format(task=repr(task_name)))
        return self.__tasks[task_name](**kwargs)



from typing import Mapping, Optional, Callable, Any

from .base import _ITask
from ..section import SectionCollection
from ...control.model import Identification, ResourceLimit
from ...utils import wrap_empty


class Task(_ITask):
    def __init__(self, name: str, identification, resources, environ, sections):
        """
        :param name: name of task
        :param identification: identification
        :param resources: resource limit
        :param environ: environment variables
        :param sections: sections
        """
        self.__name = name
        self.__identification = identification
        self.__resources = resources
        self.__environ = environ
        self.__sections = sections

        _ITask.__init__(self, self.__name, self.__identification, self.__resources, self.__environ, self.__sections)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def identification(self) -> Identification:
        return self.__identification

    @property
    def resources(self) -> ResourceLimit:
        return self.__resources

    @property
    def environ(self) -> Mapping[str, str]:
        return self.__environ

    @property
    def sections(self) -> SectionCollection:
        return self.__sections

    def __call__(self, task_start: Optional[Callable[['Task'], None]] = None,
                 task_complete: Optional[Callable[['Task', Any], None]] = None,
                 **kwargs):
        """
        run this task
        :return: return value of this task
        """
        wrap_empty(task_start)(self)
        _return = self.__sections(**kwargs)
        wrap_empty(task_complete)(self, _return)
        return _return



from typing import Union, Callable, TypeVar

from hbutils.encoding import auto_decode


def _auto_encode(s: str) -> bytes:
    return s.encode()


_T = TypeVar('_T')


def auto_encode_support(func: Callable[[Union[bytes, bytearray], ], _T]) \
        -> Callable[[Union[bytes, bytearray, str], ], _T]:
    def _func(data: Union[bytes, bytearray, str]) -> _T:
        if isinstance(data, (bytes, bytearray)):
            return func(data)
        elif isinstance(data, str):
            return func(_auto_encode(data))
        else:
            raise TypeError("Unknown type to encode support - {cls}".format(cls=type(data).__class__))

    return _func


def auto_decode_support(func: Callable[..., Union[bytes, bytearray, str]]) -> Callable[..., str]:
    def _func(*args, **kwargs) -> str:
        result = func(*args, **kwargs)
        if isinstance(result, (bytes, bytearray)):
            return auto_decode(result)
        elif isinstance(result, str):
            return result
        else:
            raise TypeError("Unknown type to decode support - {cls}".format(cls=type(result).__class__))

    return _func



import os
from abc import ABCMeta
from typing import Optional, Mapping, Callable

from hbutils.string import env_template, truncate
from pysyslimit import FilePermission

from .base import FileInputTemplate, FileInput, _load_privilege, _apply_privilege_and_identification
from ...base import _check_os_path, _check_workdir_file, _process_environ
from ....control.model import Identification
from ....utils import auto_copy_file, get_repr_info, wrap_empty


class _ICopyFileInput(metaclass=ABCMeta):
    def __init__(self, file: str, local: str, privilege, identification):
        """
        :param file: file path
        :param local: local path
        :param privilege: local privilege
        :param identification: local path identification
        """
        self.__file = file
        self.__local = local
        self.__privilege = privilege
        self.__identification = identification

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('file', lambda: repr(self.__file)),
                ('local', lambda: repr(self.__local)),
                ('privilege', lambda: repr(self.__privilege.sign), lambda: self.__privilege is not None),
                ('identification',
                 lambda: truncate(repr(self.__identification), width=48, show_length=True, tail_length=16),
                 lambda: self.__identification and self.__identification != Identification.loads({})),
            ]
        )


class CopyFileInputTemplate(FileInputTemplate, _ICopyFileInput):
    def __init__(self, file: str, local: str, privilege=None, identification=None):
        """
        :param file: file path
        :param local: local path
        :param privilege: local path privilege
        :param identification: local path identification
        """
        self.__file = file
        self.__local = local
        self.__privilege = _load_privilege(privilege)
        self.__identification = Identification.loads(identification)

        _ICopyFileInput.__init__(self, self.__file, self.__local, self.__privilege, self.__identification)

    @property
    def file(self) -> str:
        return self.__file

    @property
    def local(self) -> str:
        return self.__local

    @property
    def privilege(self) -> Optional[FilePermission]:
        return self.__privilege

    # noinspection DuplicatedCode
    def __call__(self, scriptdir: str, workdir: str, identification=None,
                 environ: Optional[Mapping[str, str]] = None, **kwargs) -> 'CopyFileInput':
        """
        generate copy file input object from extension information
        :param scriptdir: script directory
        :param workdir: work directory
        :param identification: identification
        :param environ: environment variable
        :return: copy file input object
        """
        environ = _process_environ(environ)
        _file = os.path.normpath(
            os.path.abspath(os.path.join(scriptdir, _check_os_path(env_template(self.__file, environ)))))
        _local = os.path.normpath(
            os.path.abspath(os.path.join(workdir, _check_workdir_file(env_template(self.__local, environ)))))
        _identification = Identification.merge(Identification.loads(identification), self.__identification)

        return CopyFileInput(
            file=_file, local=_local,
            privilege=self.__privilege,
            identification=_identification,
        )


class CopyFileInput(FileInput, _ICopyFileInput):
    def __init__(self, file: str, local: str,
                 privilege: Optional[FilePermission],
                 identification: Optional[Identification]):
        """
        :param file: file path
        :param local: local path
        :param privilege: local path privilege
        :param identification: identification
        """
        self.__file = file
        self.__local = local
        self.__privilege = privilege
        self.__identification = identification

        _ICopyFileInput.__init__(self, self.__file, self.__local, self.__privilege, self.__identification)

    @property
    def file(self) -> str:
        return self.__file

    @property
    def local(self) -> str:
        return self.__local

    @property
    def privilege(self) -> Optional[FilePermission]:
        return self.__privilege

    def __call__(self, input_start: Optional[Callable[['CopyFileInput'], None]] = None,
                 input_complete: Optional[Callable[['CopyFileInput'], None]] = None, **kwargs):
        """
        execute this copy event
        """
        wrap_empty(input_start)(self)
        auto_copy_file(self.__file, self.__local, self.__privilege, self.__identification.user,
                       self.__identification.group)
        _apply_privilege_and_identification(self.__local, self.__privilege, self.__identification)
        wrap_empty(input_complete)(self)



from typing import Optional, Mapping, Tuple, Any

from hbutils.string import env_template

from .base import SectionInfoTemplate, SectionInfo
from ...base import _process_environ
from ....utils import get_repr_info


class _IStaticSectionInfo:
    def __init__(self, value: str):
        """
        :param value: static value 
        """
        self.__value = value

    def __repr__(self):
        """
        :return: representation string 
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('value', lambda: repr(self.__value)),
            ]
        )


class StaticSectionInfoTemplate(SectionInfoTemplate, _IStaticSectionInfo):
    def __init__(self, value: str):
        """
        :param value: static value 
        """
        self.__value = value
        _IStaticSectionInfo.__init__(self, self.__value)

    @property
    def value(self):
        return self.__value

    def __call__(self, environ: Optional[Mapping[str, str]] = None, **kwargs) -> 'StaticSectionInfo':
        """
        get static info info object
        :param environ: environment variables
        :return: static info info object
        """
        environ = _process_environ(environ)
        if isinstance(self.__value, str):
            _value = env_template(self.__value, environ)
        else:
            _value = self.__value

        return StaticSectionInfo(value=_value)


class StaticSectionInfo(SectionInfo, _IStaticSectionInfo):
    def __init__(self, value: str):
        """
        :param value: static value 
        """
        self.__value = value
        _IStaticSectionInfo.__init__(self, self.__value)

    @property
    def value(self):
        return self.__value

    def __call__(self, **kwargs) -> Tuple[bool, Any]:
        """
        execute this info info
        """
        return True, self.__value



import shlex


def args_split(args):
    """
    Split string args into list (just keep it when list)
    :param args: args
    :return: list args
    """
    if isinstance(args, str):
        return shlex.split(args)
    else:
        return args



import click


def _click_exception_with_exit_code(name: str, message: str, exitcode: int):
    """
    create an exception
    :param name: exception name
    :param message: exception message
    :param exitcode: exitcode
    :return: new exception object
    """

    class _ClickException(click.ClickException):
        exit_code = exitcode

    return type(name, (_ClickException,), {})(message)


def _raise_exception_with_exit_code(exitcode: int, message: str = None):
    """
    raise an exception with the exit code
    :param exitcode: exitcode
    :param message: exit message
    :return:
    """
    message = message or 'exited with code {code}'.format(code=repr(exitcode))
    name = 'ExitCode{code}Exception'.format(code=repr(exitcode))
    raise _click_exception_with_exit_code(name, message, exitcode)



import json
from json import JSONDecodeError
from typing import Union

import yaml
from yaml.parser import ParserError as YamlParserError

_ERRORS_TYPING = Union[JSONDecodeError, YamlParserError]


class JsonLoadError(Exception):
    def __init__(self, exception: _ERRORS_TYPING):
        self.__exception = exception
        Exception.__init__(self, str(self.__exception))

    @property
    def exception(self) -> _ERRORS_TYPING:
        return self.__exception


_METHODS = [
    (json.load, JSONDecodeError),
    (yaml.safe_load, YamlParserError),
]


def auto_load_json(stream):
    _init_position = stream.tell()
    _last_err = None

    for _func, _except in _METHODS:
        try:
            stream.seek(_init_position)
            return _func(stream)
        except _except as err:
            _last_err = err

    raise JsonLoadError(_last_err)



import os
from typing import Callable, Mapping, Any

import click
from hbutils.scale import size_to_bytes_str
from hbutils.string import truncate

from ..event import _load_dispatch_getter as _load_abstract_dispatch_getter
from ..runner import DispatchRunner
from ...control import RunResult, RunResultStatus
from ...service import Dispatch, Command, FileInput, CopyFileInput, TagFileInput, LinkFileInput, \
    Section, SectionInfoMapping, FileOutput, CopyFileOutput, TagFileOutput


class DispatchEventRunner(DispatchRunner):
    def _section_start(self, section: Section):
        click.echo(click.style('Section {section} start ...'.format(
            section=repr(section.name)
        ), fg='blue'))

    def _section_complete(self, section: Section, result):
        _success, _results, _info = result
        _color = 'green' if _success else 'red'
        _result_str = 'completed' if _success else 'failed'

        click.echo(click.style('Section {section} execute {result}!'.format(
            section=repr(section.name),
            result=_result_str,
        ), fg=_color))
        click.echo('')

    def _input_start(self, input_: FileInput):
        if isinstance(input_, CopyFileInput):
            type_ = 'directory' if os.path.isdir(input_.file) else 'file'
            _sentence = "Coping {type} from {from_} to {to} ... ".format(type=type_, from_=repr(input_.file),
                                                                         to=repr(input_.local))

        elif isinstance(input_, TagFileInput):
            _sentence = "Loading tag {tag} to {to} ... ".format(tag=repr(input_.tag), to=repr(input_.local))
        elif isinstance(input_, LinkFileInput):
            type_ = 'directory' if os.path.isdir(input_.file) else 'file'
            _sentence = "Linking {type} from {from_} to {to} ... ".format(type=type_, from_=repr(input_.file),
                                                                          to=repr(input_.local))
        else:
            raise TypeError('Invalid file input object - {repr}.'.format(repr=repr(input_)))

        click.echo(click.style(_sentence, bold=False), nl=False)

    def _input_complete(self, input_: FileInput):
        click.echo(click.style('COMPLETE', fg='green'))

    def _command_start(self, command: Command):
        _repr = truncate(repr(command.args), width=96, tail_length=24, show_length=True)
        click.echo(click.style("Running {repr} ... ".format(repr=_repr), bold=True), nl=False)

    def _command_complete(self, command: Command, result: RunResult):
        _color = 'green' if result.ok else 'red'
        if result.status != RunResultStatus.NOT_COMPLETED:
            _title = '{status}, time: {cpu_time} / {real_time}, memory: {memory}'.format(
                status=result.status.name,
                cpu_time='%.3fs' % result.result.cpu_time,
                real_time='%.3fs' % result.result.real_time,
                memory=size_to_bytes_str(result.result.max_memory),
            )
        else:
            _title = result.status.name
        click.echo(click.style(_title, fg=_color, bold=True), nl=True)

    def _output_start(self, output: FileOutput):
        if isinstance(output, CopyFileOutput):
            type_ = 'directory' if os.path.isdir(output.local) else 'file'
            _sentence = "Coping {type} from {from_} to {to} ... ".format(type=type_, from_=repr(output.local),
                                                                         to=repr(output.file))
        elif isinstance(output, TagFileOutput):
            type_ = 'directory' if os.path.isdir(output.local) else 'file'
            _sentence = "Saving {type} from {from_} to tag {tag} ... ".format(type=type_, tag=repr(output.tag),
                                                                              from_=repr(output.local))
        else:
            raise TypeError('Invalid file output object - {repr}.'.format(repr=repr(output)))

        click.echo(click.style(_sentence, bold=False), nl=False)

    def _output_complete(self, output: FileOutput):
        click.echo(click.style('COMPLETE', fg='green'))

    def _info_mapping_start(self, mapping: SectionInfoMapping):
        click.echo(click.style('Collecting result information ... ', bold=False), nl=False)

    def _info_mapping_complete(self, mapping: SectionInfoMapping, result):
        click.echo(click.style('COMPLETE', fg='green'), nl=True)

    def _info_dump_start(self, section: Section, info: Mapping[str, Any]):
        click.echo(click.style('Dumping result information to {dump} ... '.format(
            dump=repr(section.info_dump)), bold=False), nl=False)

    def _info_dump_complete(self, section: Section, info: Mapping[str, Any]):
        click.echo(click.style('COMPLETE', fg='green'), nl=True)


def _load_dispatch_getter(filename: str = None) -> Callable[..., Dispatch]:
    return _load_abstract_dispatch_getter(DispatchEventRunner, filename)



import os
from typing import Mapping, Optional, Any

from hbutils.string import env_template

from ..utils import is_inner_relative_path, FilePool


def _check_os_path(path: str) -> str:
    """
    check file valid or not, when valid, just process it
    :param path: original file path
    :return: normalized file path
    """
    return os.path.normpath(path)


def _check_workdir_position(path: str) -> str:
    """
    check local path valid or not, when valid, just process it
    :param path: original local path
    :return: normalized local path
    """
    if not is_inner_relative_path(path):
        raise ValueError(
            'Inner relative position expected for local but {actual} found.'.format(actual=repr(path)))
    return os.path.normpath(path)


def _check_workdir_file(path: str) -> str:
    """
    check local path valid or not, when valid, just process it
    :param path: original local path
    :return: normalized local path
    """
    if not is_inner_relative_path(path, allow_root=False):
        raise ValueError(
            'Inner relative file path expected for local but {actual} found.'.format(actual=repr(path)))
    return os.path.normpath(path)


def _check_pool_tag(tag: str) -> str:
    """
    check if tag is valid, if valid, just return it
    :param tag: tag
    :return: original tag
    """
    FilePool.check_tag_name(tag)
    return tag


def _process_environ(environ: Optional[Mapping[str, Any]] = None,
                     ext_environ: Optional[Mapping[str, Any]] = None,
                     enable_ext: bool = False) -> Mapping[str, str]:
    """
    process environment variables
    :param environ: environment variables
    :param ext_environ: external environment variables
    :param enable_ext: enable external
    :return: string environment variables
    """

    def _penv(env):
        return {key: str(value) for key, value in (env or {}).items()}

    _curenv = _penv(environ)
    _extenv = _penv(ext_environ)
    if enable_ext:
        _curenv = {key: env_template(value, _extenv) for key, value in _curenv.items()}

    _result = dict(_extenv)
    _result.update(**_curenv)

    return _result



import os
from abc import ABCMeta
from typing import Optional, Mapping

from hbutils.string import truncate

from .global_ import GlobalConfigTemplate, GlobalConfig
from ..task import TaskMappingTemplate, TaskMapping
from ...utils import get_repr_info


class _IDispatch(metaclass=ABCMeta):
    def __init__(self, global_, tasks):
        """
        :param global_: global config
        :param tasks: task mapping
        """
        self.__global = global_
        self.__tasks = tasks

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('tasks', lambda: truncate(repr(tuple(sorted(self.__tasks.items.keys()))),
                                           width=64, show_length=True, tail_length=16),
                 lambda: self.__tasks),
            ]
        )


_KEY_MAPPING = {'global': 'global_'}


class DispatchTemplate(_IDispatch):
    def __init__(self, tasks, global_=None):
        """
        :param global_: global config template
        :param tasks: task mapping template
        """
        self.__global = GlobalConfigTemplate.loads(global_ or {})
        self.__tasks = TaskMappingTemplate.loads(tasks)
        _IDispatch.__init__(self, self.__global, self.__tasks)

    @property
    def global_(self) -> GlobalConfigTemplate:
        return self.__global

    @property
    def tasks(self) -> TaskMappingTemplate:
        return self.__tasks

    def __call__(self, scriptdir: str = None, environ: Optional[Mapping[str, str]] = None,
                 environ_after: Optional[Mapping[str, str]] = None, **kwargs):
        """
        generate dispatch object
        :param scriptdir: script directory (default is '.')
        :param environ: environment variable
        :param environ_after: force replace environment variable
        :param kwargs: other arguments
        :return: dispatch object
        """
        scriptdir = scriptdir or os.path.abspath(os.curdir)

        _global = self.__global(environ, environ_after, **kwargs)
        _tasks = self.__tasks(
            scriptdir=scriptdir,
            identification=_global.identification,
            resources=_global.resources,
            environ=_global.environ, **kwargs
        )

        return Dispatch(global_=_global, tasks=_tasks)

    @classmethod
    def loads(cls, data) -> 'DispatchTemplate':
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls(**{_KEY_MAPPING.get(key, key): value for key, value in data.items()})
        else:
            raise TypeError('Json or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))


class Dispatch(_IDispatch):
    def __init__(self, global_, tasks):
        """
        :param global_: global config
        :param tasks: task mapping
        """
        self.__global = global_
        self.__tasks = tasks
        _IDispatch.__init__(self, self.__global, self.__tasks)

    @property
    def global_(self) -> GlobalConfig:
        return self.__global

    @property
    def tasks(self) -> TaskMapping:
        return self.__tasks

    def __call__(self, task_name: str, **kwargs):
        """
        run a task
        :param task_name: name of a task
        :return: result of task
        """
        return self.__tasks(task_name, **kwargs)



from enum import unique, IntEnum
from multiprocessing import Lock
from typing import Optional

from .process import ProcessResult
from .resource import ResourceLimit
from ...utils import get_repr_info


@unique
class RunResultStatus(IntEnum):
    NOT_COMPLETED = -1
    SUCCESS = 0
    CPU_TIME_LIMIT_EXCEED = 1
    REAL_TIME_LIMIT_EXCEED = 2
    MEMORY_LIMIT_EXCEED = 3
    RUNTIME_ERROR = 4
    SYSTEM_ERROR = 5

    @property
    def ok(self):
        """
        :return: run success or not
        """
        return self == RunResultStatus.SUCCESS

    @property
    def completed(self):
        """
        :return: process completed or not
        """
        return self != RunResultStatus.NOT_COMPLETED


class RunResult:
    def __init__(self, limit: ResourceLimit, result: Optional[ProcessResult]):
        """
        :param limit: resource limit
        :param result: process running result
        """
        self.__limit = limit
        self.__result = result
        self.__lock = Lock()

    def __get_status(self) -> RunResultStatus:
        """
        Get status information
        :return: status information
        """
        if self.__result is None:
            return RunResultStatus.NOT_COMPLETED
        elif self.__limit.max_cpu_time is not None and self.__result.cpu_time > self.__limit.max_cpu_time:
            return RunResultStatus.CPU_TIME_LIMIT_EXCEED
        elif self.__limit.max_real_time is not None and self.__result.real_time > self.__limit.max_real_time:
            return RunResultStatus.REAL_TIME_LIMIT_EXCEED
        elif self.__limit.max_memory is not None and self.__result.max_memory > self.__limit.max_memory:
            return RunResultStatus.MEMORY_LIMIT_EXCEED
        elif self.__result.exitcode != 0:
            return RunResultStatus.RUNTIME_ERROR
        elif not self.__result.ok:
            return RunResultStatus.SYSTEM_ERROR
        else:
            return RunResultStatus.SUCCESS

    @property
    def limit(self) -> ResourceLimit:
        """
        :return: resource limit
        """
        return self.__limit

    @property
    def result(self) -> ProcessResult:
        """
        :return: process running result
        """
        with self.__lock:
            return self.__result

    @property
    def status(self) -> RunResultStatus:
        """
        :return: current status
        """
        with self.__lock:
            return self.__get_status()

    @property
    def ok(self) -> bool:
        """
        :return: process run success or not
        """
        with self.__lock:
            return self.__result is not None and self.__result.ok and self.__get_status().ok

    @property
    def completed(self) -> bool:
        """
        :return: process run completed or not
        """
        with self.__lock:
            return self.__get_status().completed

    def __repr__(self):
        """
        :return: string representation format
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('status', lambda: self.status.name),
                ('exitcode', (lambda: self.result.exitcode, lambda: self.result.exitcode != 0)),
                ('signal', (lambda: self.result.signal.name, lambda: self.result.signal is not None)),
            ],
        )

    @property
    def json(self):
        """
        :return: get run result information
        """
        return {
            'limit': self.limit.json,
            'result': self.result.json if self.result else None,
            'status': self.status.name,
            'ok': self.ok,
            'completed': self.completed,
        }



import re
from abc import ABCMeta

from hbutils.string import truncate

from ....control.model import Identification, ResourceLimit
from ....utils import get_repr_info

_SECTION_NAME_PATTERN = re.compile('[a-zA-Z_][0-9a-zA-Z_]*')


def _check_section_name(name: str) -> str:
    """
    check section name valid or not
    :param name: section name
    :return: section name
    """
    if _SECTION_NAME_PATTERN.fullmatch(name):
        return name
    else:
        raise ValueError('Section name should match {pattern} but {actual} found.'.format(
            pattern=repr(_SECTION_NAME_PATTERN.pattern),
            actual=repr(name),
        ))


ENV_PJI_SECTION_NAME = 'PJI_SECTION_NAME'


class _ISection(metaclass=ABCMeta):
    def __init__(self, name: str, identification, resources, environ,
                 inputs, outputs, infos, commands):
        """
        :param name: name of section
        :param identification: identification used
        :param resources: resource limits
        :param environ: environment variables
        :param inputs: inputs
        :param outputs: outputs
        :param infos: infos
        :param commands: commands
        """
        self.__name = name
        self.__identification = identification
        self.__resources = resources
        self.__environ = environ

        self.__inputs = inputs
        self.__outputs = outputs
        self.__infos = infos
        self.__commands = commands

    def __repr__(self):
        """
        :return: get representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('name', lambda: repr(self.__name)),
                ('identification',
                 lambda: truncate(repr(self.__identification), width=48, show_length=True, tail_length=16),
                 lambda: self.__identification and self.__identification != Identification.loads({})),
                ('resources', lambda: truncate(repr(self.__resources), width=64, show_length=True, tail_length=16),
                 lambda: self.__resources != ResourceLimit.loads({})),
                ('inputs', lambda: repr(len(self.__inputs.items)),
                 lambda: self.__inputs and len(self.__inputs.items) > 0),
                ('outputs', lambda: repr(len(self.__outputs.items)),
                 lambda: self.__outputs and len(self.__outputs.items) > 0),
                ('infos', lambda: repr(len(self.__infos.items)),
                 lambda: self.__infos and len(self.__infos.items.keys()) > 0),
                ('commands', lambda: repr(len(self.__commands.commands)),
                 lambda: self.__commands and len(self.__commands.commands) > 0),
            ]
        )



import os
from functools import wraps
from typing import Optional

import where

from ..model import ResourceLimit, Identification
from ...utils import args_split


def _do_nothing():
    pass


def _attach_preexec_fn(preexec_fn=None, pre_attach=None, post_attach=None):
    preexec_fn = preexec_fn or _do_nothing
    pre_attach = pre_attach or _do_nothing
    post_attach = post_attach or _do_nothing

    @wraps(preexec_fn)
    def _new_preexec_fn():
        pre_attach()
        preexec_fn()
        post_attach()

    return _new_preexec_fn


def workdir_setter(func):
    @wraps(func)
    def _func(*args, cwd: Optional[str] = None, preexec_fn=None, **kwargs):
        cwd = cwd or os.getcwd()
        if not os.path.exists(cwd):
            raise FileNotFoundError('Path {cwd} not found.'.format(cwd=repr(cwd)))
        if not os.path.isdir(cwd):
            raise NotADirectoryError('{cwd} is not a directory.'.format(cwd=repr(cwd)))

        def _change_dir_func():
            os.chdir(cwd)

        preexec_fn = _attach_preexec_fn(preexec_fn, pre_attach=_change_dir_func)
        return func(*args, preexec_fn=preexec_fn, **kwargs)

    return _func


_REAL_TIME_LIMIT_KEY = 'real_time_limit'


def resources_setter(func):
    @wraps(func)
    def _func(*args, resources=None, preexec_fn=None, **kwargs):
        resources = ResourceLimit.loads(resources)

        def _apply_resource_limit_func():
            resources.apply()

        preexec_fn = _attach_preexec_fn(preexec_fn, post_attach=_apply_resource_limit_func)
        return func(*args, preexec_fn=preexec_fn, resources=resources, **kwargs)

    return _func


def users_setter(func):
    @wraps(func)
    def _func(*args, identification=None, preexec_fn=None, **kwargs):
        identification = Identification.loads(identification)

        def _apply_user_func():
            identification.apply()

        preexec_fn = _attach_preexec_fn(preexec_fn, post_attach=_apply_user_func)
        return func(*args, preexec_fn=preexec_fn, **kwargs)

    return _func


def shell_setter(func):
    @wraps(func)
    def _func(*args_, args, shell: bool = False, **kwargs):
        if shell:
            if isinstance(args, str):
                if where.first('sh'):
                    args = [where.first('sh'), '-c', args]
                elif where.first('cmd'):
                    args = [where.first('cmd'), '/c', args]
                else:
                    raise EnvironmentError('Neither shell nor cmd found in this environment.')
            else:
                raise ValueError(
                    'When shell is enabled, args should be str but {actual} found.'.format(actual=repr(type(args))))
        else:
            args = args_split(args)

        return func(*args_, args=args, **kwargs)

    return _func


def process_setter(func):
    @wraps(func)
    @users_setter
    @resources_setter
    @workdir_setter
    @shell_setter
    def _func(*args, **kwargs):
        return func(*args, **kwargs)

    return _func



import os
from abc import ABCMeta
from typing import Optional, Mapping, Callable

from hbutils.string import env_template, truncate
from pysyslimit import FilePermission

from .base import FileInput, FileInputTemplate, _load_privilege, _apply_privilege_and_identification
from ...base import _check_workdir_file, _check_pool_tag, _process_environ
from ....control.model import Identification
from ....utils import get_repr_info, FilePool, wrap_empty


class _ITagFileInput(metaclass=ABCMeta):
    def __init__(self, tag: str, local: str, privilege, identification):
        """
        :param tag: pool tag
        :param local: local path
        :param privilege: local path privilege
        :param identification: local path identification
        """
        self.__tag = tag
        self.__local = local
        self.__privilege = privilege
        self.__identification = identification

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('tag', lambda: repr(self.__tag)),
                ('local', lambda: repr(self.__local)),
                ('privilege', lambda: repr(self.__privilege.sign), lambda: self.__privilege is not None),
                ('identification',
                 lambda: truncate(repr(self.__identification), width=48, show_length=True, tail_length=16),
                 lambda: self.__identification and self.__identification != Identification.loads({})),
            ]
        )


class TagFileInputTemplate(FileInputTemplate, _ITagFileInput):
    def __init__(self, tag: str, local: str, privilege=None, identification=None):
        """
        :param tag: pool tag
        :param local: local path
        :param privilege: local path privilege
        :param identification: local path identification
        """
        self.__tag = tag
        self.__local = local
        self.__privilege = _load_privilege(privilege)
        self.__identification = Identification.loads(identification)

        _ITagFileInput.__init__(self, self.__tag, self.__local, self.__privilege, self.__identification)

    @property
    def tag(self) -> str:
        return self.__tag

    @property
    def local(self) -> str:
        return self.__local

    @property
    def privilege(self) -> Optional[FilePermission]:
        return self.__privilege

    def __call__(self, workdir: str, pool: FilePool, identification=None,
                 environ: Optional[Mapping[str, str]] = None, **kwargs) -> 'TagFileInput':
        """
        get tag file input object
        :param workdir: local work directory
        :param pool: file pool object
        :param environ: environment variables
        :return: tag file input object
        """
        environ = _process_environ(environ)
        _tag = _check_pool_tag(env_template(self.__tag, environ))
        _local = os.path.normpath(
            os.path.abspath(os.path.join(workdir, _check_workdir_file(env_template(self.__local, environ)))))
        _identification = Identification.merge(Identification.loads(identification), self.__identification)

        return TagFileInput(
            pool=pool, tag=_tag, local=_local,
            privilege=self.__privilege,
            identification=_identification,
        )


class TagFileInput(FileInput, _ITagFileInput):
    def __init__(self, pool: FilePool, tag: str, local: str,
                 privilege: Optional[FilePermission],
                 identification: Optional[Identification]):
        """
        :param pool: file pool
        :param tag: pool tag
        :param local: local path
        :param privilege: local path privilege
        :param identification: local path identification
        """
        self.__pool = pool
        self.__tag = tag
        self.__local = local
        self.__privilege = privilege
        self.__identification = identification

        _ITagFileInput.__init__(self, self.__tag, self.__local, self.__privilege, self.__identification)

    @property
    def tag(self) -> str:
        return self.__tag

    @property
    def local(self) -> str:
        return self.__local

    @property
    def privilege(self) -> Optional[FilePermission]:
        return self.__privilege

    def __call__(self, input_start: Optional[Callable[['TagFileInput'], None]] = None,
                 input_complete: Optional[Callable[['TagFileInput'], None]] = None, **kwargs):
        """
        execute this file input
        """
        wrap_empty(input_start)(self)
        self.__pool.export(self.__tag, self.__local, self.__privilege, self.__identification.user,
                           self.__identification.group)
        _apply_privilege_and_identification(self.__local, self.__privilege, self.__identification)
        wrap_empty(input_complete)(self)



from .args import args_split
from .collection import duplicates
from .context import eclosing
from .decorator import allow_none
from .encoding import auto_decode_support, auto_encode_support
from .file import FilePool, auto_copy_file, auto_link_file
from .func import wrap_empty
from .iter import gen_lock
from .json import auto_load_json, JsonLoadError
from .path import is_absolute_path, is_relative_path, is_inner_relative_path, makedirs
from .repr import get_repr_info
from .value import ValueProxy



import os
import signal
from resource import struct_rusage
from typing import Optional

from hbutils.scale import size_to_bytes_str

from ...utils import get_repr_info


class _IStatus:
    def __init__(self, status: int):
        self.__status = status

    @property
    def exitcode(self) -> int:
        """
        :return: exitcode of process
        """
        return os.WEXITSTATUS(self.__status)

    @property
    def signal_code(self) -> int:
        """
        :return: signal code of process
        """
        return os.WTERMSIG(self.__status)

    # noinspection PyArgumentList
    @property
    def signal(self) -> Optional[signal.Signals]:
        """
        :return: signal object
        """
        if self.signal_code:
            return signal.Signals(self.signal_code)
        else:
            return None

    @property
    def ok(self) -> bool:
        """
        :return: true if quit normally, otherwise false
        """
        return not self.__status


class _IDuration:
    def __init__(self, start_time: float, end_time: float):
        """
        :param start_time: process start time
        :param end_time: process end time
        """
        self.__start_time = start_time
        self.__end_time = end_time

    @property
    def start_time(self):
        """
        :return: start time of process
        """
        return self.__start_time

    @property
    def end_time(self):
        """
        :return: end time of process
        """
        return self.__end_time

    @property
    def real_time(self):
        """
        :return: duration of process
        """
        if self.start_time is not None and self.end_time is not None:
            return self.end_time - self.start_time
        else:
            return None


class _IResource:
    def __init__(self, resource_usage: struct_rusage):
        self.__resource_usage = resource_usage

    @property
    def resource_usage(self):
        """
        获取资源占用对象
        :return: 资源占用对象
        """
        return self.__resource_usage

    @property
    def cpu_time(self):
        """
        获取cpu时间使用（单位：秒）
        :return: cpu时间使用（单位：秒）
        """
        return self.__resource_usage.ru_utime

    @property
    def system_time(self):
        """
        获取系统时间使用（单位：秒）
        :return: 系统时间使用（单位：秒）
        """
        return self.__resource_usage.ru_stime

    @property
    def max_memory(self):
        """
        获取最大内存占用（单位：字节）
        :return: 最大内存占用（单位：字节）
        """
        return self.__resource_usage.ru_maxrss * 1024.0


class ProcessResult(_IStatus, _IDuration, _IResource):
    def __init__(self, status, start_time, end_time, resource_usage):
        """
        :param status: result status
        :param start_time: start time of process
        :param end_time: end time of process
        :param resource_usage: resource usage
        """
        _IStatus.__init__(self, status)
        _IDuration.__init__(self, start_time, end_time)
        _IResource.__init__(self, resource_usage)

    def __repr__(self):
        """
        :return: get presentation format
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('exitcode', lambda: self.exitcode),
                ('signal', (lambda: self.signal.name, lambda: self.signal)),
                ('real time', lambda: '%.3fs' % self.real_time),
                ('cpu time', lambda: '%.3fs' % self.cpu_time),
                ('max memory', lambda: size_to_bytes_str(self.max_memory)),
            ]
        )

    @property
    def json(self):
        """
        get process information
        :return: process information json
        """
        return {
            'exitcode': self.exitcode,
            'signal': self.signal.name if self.signal else None,
            'real_time': self.real_time,
            'cpu_time': self.cpu_time,
            'max_memory': self.max_memory,
        }



from enum import IntEnum, unique
from typing import Mapping, Any

from .base import FileOutputTemplate
from .copy import CopyFileOutputTemplate
from .tag import TagFileOutputTemplate


@unique
class FileOutputType(IntEnum):
    COPY = 1
    TAG = 2

    @classmethod
    def loads(cls, value) -> 'FileOutputType':
        """
        Load FileOutputType from value
        :param value: raw value
        :return: file output type object
        """
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            if value.upper() in cls.__members__.keys():
                return cls.__members__[value.upper()]
            else:
                raise KeyError('Unknown file output type - {actual}.'.format(actual=repr(value)))
        elif isinstance(value, int):
            _mapping = {v.value: v for k, v in cls.__members__.items()}
            if value in _mapping.keys():
                return _mapping[value]
            else:
                raise ValueError('Unknown file output type value - {actual}'.format(actual=repr(value)))
        else:
            raise TypeError('Int, str or {cls} expected but {actual} found.'.format(
                cls=cls.__name__,
                actual=repr(type(value).__name__)
            ))


_TYPE_TO_TEMPLATE_CLASS = {
    FileOutputType.COPY: CopyFileOutputTemplate,
    FileOutputType.TAG: TagFileOutputTemplate,
}


# noinspection DuplicatedCode
def _load_output_template_from_json(json: Mapping[str, Any]) -> FileOutputTemplate:
    """
    load template object from json data
    :param json: json data
    :return: file output template object
    """
    if 'type' not in json.keys():
        raise KeyError('Key {type} not found.'.format(type=repr('type')))

    _type = FileOutputType.loads(json['type'])
    _json = dict(json)
    del _json['type']

    return _TYPE_TO_TEMPLATE_CLASS[_type](**_json)


def _load_output_template_from_string(string: str) -> FileOutputTemplate:
    """
    load template object from string (should split with :)
    :param string: string value
    :return: file output template object
    """
    _items = string.split(':')
    _type, _args = FileOutputType.loads(_items[0].strip()), [_item.strip() for _item in _items[1:]]
    return _TYPE_TO_TEMPLATE_CLASS[_type](*_args)


def load_output_template(data) -> FileOutputTemplate:
    """
    load file output template object from data
    :param data: raw data
    :return: file output template object
    """
    if isinstance(data, FileOutputTemplate):
        return data
    elif isinstance(data, dict):
        return _load_output_template_from_json(data)
    elif isinstance(data, str):
        return _load_output_template_from_string(data)
    else:
        raise TypeError('Json, string or {type} expected but {actual} found.'.format(
            type=FileOutputTemplate.__name__, actual=repr(type(data).__name__)))



from typing import List, Optional, Callable

from .base import FileOutputTemplate, FileOutput
from .general import load_output_template
from ....utils import get_repr_info, wrap_empty


class _IFileOutputCollection:
    def __init__(self, items):
        """
        :param items: file outputs
        """
        self.__items = items

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('outputs', lambda: len(self.__items)),
            ]
        )


class FileOutputCollectionTemplate(_IFileOutputCollection):
    def __init__(self, *items):
        """
        :param items: file output templates
        """
        self.__items = [load_output_template(item) for item in items]
        _IFileOutputCollection.__init__(self, self.__items)

    @property
    def items(self) -> List[FileOutputTemplate]:
        return list(self.__items)

    def __iter__(self):
        return self.items.__iter__()

    def __call__(self, **kwargs):
        """
        generate file output collection
        :param kwargs: plenty of arguments
        :return: file output collection
        """
        return FileOutputCollection(*[item(**kwargs) for item in self.__items])

    @classmethod
    def loads(cls, data) -> 'FileOutputCollectionTemplate':
        """
        load file output collection template from data
        :param data: raw data
        :return: file output collection template
        """
        data = data or []
        if isinstance(data, cls):
            return data
        elif isinstance(data, FileOutputTemplate):
            return cls(data)
        elif isinstance(data, (list, tuple)):
            return cls(*data)
        elif isinstance(data, (dict, str)):
            return cls(load_output_template(data))
        else:
            raise TypeError('Array or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))


class FileOutputCollection(_IFileOutputCollection):
    def __init__(self, *items):
        """
        :param items: file outputs
        """
        self.__items = list(items)
        _IFileOutputCollection.__init__(self, self.__items)

    @property
    def items(self) -> List[FileOutput]:
        return list(self.__items)

    def __iter__(self):
        return self.items.__iter__()

    def __call__(self, output_collection_start: Optional[Callable[['FileOutputCollection'], None]] = None,
                 output_collection_complete: Optional[Callable[['FileOutputCollection'], None]] = None, **kwargs):
        """
        execute this file output setting
        """
        wrap_empty(output_collection_start)(self)
        for item in self.__items:
            item(**kwargs)
        wrap_empty(output_collection_complete)(self)



import os
from enum import unique, IntEnum
from typing import Union, List, Optional

from hbutils.string import truncate

from ...control.model import ResourceLimit, Identification
from ...utils import get_repr_info


@unique
class CommandMode(IntEnum):
    COMMON = 1
    TIMING = 2
    MUTUAL = 3

    @classmethod
    def loads(cls, value) -> 'CommandMode':
        """
        Load CommandMode from value
        :param value: raw value
        :return: CommandMode object
        """
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            if value.upper() in cls.__members__.keys():
                return cls.__members__[value.upper()]
            else:
                raise KeyError('Unknown command mode - {actual}.'.format(actual=repr(value)))
        elif isinstance(value, int):
            _mapping = {v.value: v for k, v in cls.__members__.items()}
            if value in _mapping.keys():
                return _mapping[value]
            else:
                raise ValueError('Unknown command mode value - {actual}'.format(actual=repr(value)))
        else:
            raise TypeError('Int, str or {cls} expected but {actual} found.'.format(
                cls=cls.__name__,
                actual=repr(type(value).__name__)
            ))


ENV_PJI_COMMAND_INDEX = 'PJI_COMMAND_INDEX'


class _ICommandBase:
    def __init__(self, args: Union[str, List[str]], shell: bool = True, workdir: Optional[str] = None,
                 identification=None, resources=None, mode=None):
        """
        :param args: arguments
        :param shell: use shell mode
        :param workdir: work directory
        :param identification: identification used
        :param resources: resource limits
        :param mode: command mode value
        """
        self.__args = args
        self.__shell = shell
        self.__workdir = workdir
        self.__identification = identification
        self.__resources = resources
        self.__mode = mode

    def __repr__(self):
        """
        :return: get representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('args', lambda: truncate(repr(self.__args), width=48, show_length=True, tail_length=16)),
                ('shell', lambda: repr(self.__shell)),
                ('mode', lambda: self.__mode.name),
                ('workdir', lambda: repr(self.__workdir),
                 lambda: os.path.normpath(self.__workdir) != os.path.normpath('.')),
                ('identification',
                 lambda: truncate(repr(self.__identification), width=48, show_length=True, tail_length=16),
                 lambda: self.__identification and self.__identification != Identification.loads({})),
                ('resources', lambda: truncate(repr(self.__resources), width=64, show_length=True, tail_length=16),
                 lambda: self.__resources and self.__resources != ResourceLimit.loads({})),
            ]
        )



from functools import reduce
from typing import Mapping, Union, Optional

from pysyslimit import SystemUser, SystemGroup

from ...utils import get_repr_info


class Identification:
    def __init__(self, user=None, group=None, auto_group: bool = False):
        if user is not None:
            self.__user = SystemUser.loads(user)
        else:
            self.__user = None

        if group is not None:
            self.__group = SystemGroup.loads(group)
        elif auto_group and self.__user is not None:
            self.__group = self.__user.primary_group
        else:
            self.__group = None

    @property
    def user(self) -> SystemUser:
        return self.__user

    @property
    def group(self) -> SystemGroup:
        return self.__group

    def apply(self):
        if self.__group is not None:
            self.__group.apply()
        if self.__user is not None:
            self.__user.apply(include_group=False)

    def to_json(self) -> Mapping[str, Optional[Union[SystemUser, SystemGroup]]]:
        return {
            'user': self.user.name if self.user else None,
            'group': self.group.name if self.group else None,
        }

    @classmethod
    def current(cls) -> 'Identification':
        return cls.loads((SystemUser.current(), SystemGroup.current()))

    @classmethod
    def load_from_file(cls, filename: str) -> 'Identification':
        return cls.loads((SystemUser.load_from_file(filename),
                          SystemGroup.load_from_file(filename)))

    @classmethod
    def loads(cls, data) -> 'Identification':
        data = data or {}
        if isinstance(data, Identification):
            return data
        elif isinstance(data, dict):
            return cls(
                user=data.get('user', None),
                group=data.get('group', None),
                auto_group=False,
            )
        elif isinstance(data, tuple):
            _user, _group = data
            return cls(user=_user, group=_group, auto_group=False)
        elif isinstance(data, SystemUser):
            return cls(user=data, auto_group=True)
        elif isinstance(data, SystemGroup):
            return cls(user=None, group=data)
        else:
            try:
                return cls(user=SystemUser.loads(data), auto_group=True)
            except KeyError:
                raise ValueError('Unrecognized {actual} value for {cls}.'.format(
                    actual=repr(type(data)),
                    cls=repr(cls),
                ))

    @classmethod
    def merge(cls, *commands) -> 'Identification':
        def _merge(a: 'Identification', b: 'Identification') -> 'Identification':
            return cls(
                user=b.user or a.user,
                group=b.group or a.group,
            )

        # noinspection PyTypeChecker
        commands = [cls.loads(item) for item in commands]
        return reduce(_merge, commands, cls())

    def __tuple(self):
        _json = self.to_json()
        return _json['user'], _json['group']

    def __hash__(self):
        return hash(self.__tuple())

    def __eq__(self, other) -> bool:
        if other is self:
            return True
        elif isinstance(other, Identification):
            return other.to_json() == self.to_json()
        else:
            return False

    def __repr__(self) -> str:
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('user', (lambda: self.user.name, lambda: self.user)),
                ('group', (lambda: self.group.name, lambda: self.group)),
            ]
        )



from abc import ABCMeta
from typing import List, Tuple, Optional, Callable

from .base import ENV_PJI_COMMAND_INDEX
from .command import Command
from .template import CommandTemplate
from ..base import _process_environ
from ...control.model import RunResult
from ...utils import get_repr_info, wrap_empty


class _ICommandCollection(metaclass=ABCMeta):
    def __init__(self, commands):
        """
        :param commands: commands
        """
        self.__commands = commands

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('commands', lambda: len(self.__commands))
            ]
        )


class CommandCollectionTemplate(_ICommandCollection):
    def __init__(self, *commands: CommandTemplate):
        """
        :param commands: tuple of command templates
        """
        self.__commands = [CommandTemplate.loads(item) for item in commands]

        _ICommandCollection.__init__(self, self.__commands)

    @property
    def commands(self) -> List[CommandTemplate]:
        return list(self.__commands)

    def __call__(self, identification=None, resources=None, workdir=None, environ=None,
                 **kwargs) -> 'CommandCollection':
        """
        generate command collection
        :param identification: identification
        :param resources: resource limits
        :param workdir: work directory
        :param environ: environment variables
        :return: command collection
        """
        environ = _process_environ(environ)

        def _env_with_id(index_):
            _env = dict(environ)
            _env[ENV_PJI_COMMAND_INDEX] = str(index_)
            return _env

        return CommandCollection(*[item(
            identification, resources, workdir, _env_with_id(index), **kwargs
        ) for index, item in enumerate(self.__commands)])

    @classmethod
    def loads(cls, data) -> 'CommandCollectionTemplate':
        """
        load command collection template from data
        :param data: raw data
        :return: command collection template object
        """
        if isinstance(data, cls):
            return data
        elif isinstance(data, CommandTemplate):
            return cls(data)
        elif isinstance(data, (list, tuple)):
            return cls(*data)
        elif isinstance(data, dict):
            return cls(CommandTemplate.loads(data))
        else:
            raise TypeError('List or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))


_COLLECTION_RESULT = Tuple[bool, List[RunResult]]


class CommandCollection(_ICommandCollection):
    def __init__(self, *commands: Command):
        """
        :param commands: tuple of commands
        """
        self.__commands = commands

        _ICommandCollection.__init__(self, self.__commands)

    @property
    def commands(self) -> List[Command]:
        return list(self.__commands)

    def __call__(self, command_collection_start: Optional[Callable[['CommandCollection'], None]] = None,
                 command_collection_complete: Optional[
                     Callable[['CommandCollection', _COLLECTION_RESULT], None]] = None,
                 **kwargs) -> _COLLECTION_RESULT:
        """
        execute multiple commands one by one
        :return: success or not, list of results
        """
        wrap_empty(command_collection_start)(self)
        _results = []
        _success = True
        for index, cmd in enumerate(self.__commands):
            result = cmd(**kwargs)
            _results.append(result)
            if not result.ok:
                _success = False
                break

        _return = (_success, _results)
        wrap_empty(command_collection_complete)(self, _return)
        return _return



import threading


class _TLockedIterator(object):
    def __init__(self, generator):
        self.__lock = threading.Lock()
        self.__generator = iter(generator)

    def __iter__(self):
        return self

    def __next__(self):
        with self.__lock:
            return next(self.__generator)


def gen_lock(gen):
    return _TLockedIterator(gen)



def wrap_empty(func):
    """
    wrap a function which can be None
    :param func: optional function
    :return: final function
    """
    if func:
        return func
    else:
        def _empty_func(*args, **kwargs):
            pass

        return _empty_func



import os
from abc import ABCMeta
from typing import Optional, Mapping, Callable

from hbutils.string import env_template

from .base import FileOutputTemplate, FileOutput
from ...base import _check_workdir_file, _check_pool_tag, _process_environ
from ....utils import get_repr_info, FilePool, wrap_empty


class _ITagFileOutput(metaclass=ABCMeta):
    def __init__(self, local: str, tag: str):
        """
        :param local: local path
        :param tag: pool tag
        """
        self.__local = local
        self.__tag = tag

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('local', lambda: repr(self.__local)),
                ('tag', lambda: repr(self.__tag)),
            ]
        )


class TagFileOutputTemplate(FileOutputTemplate, _ITagFileOutput):
    def __init__(self, local: str, tag: str):
        """
        :param local: local path
        :param tag: pool tag
        """
        self.__local = local
        self.__tag = tag

        _ITagFileOutput.__init__(self, self.__local, self.__tag)

    @property
    def tag(self) -> str:
        return self.__tag

    @property
    def local(self) -> str:
        return self.__local

    def __call__(self, workdir: str, pool: FilePool,
                 environ: Optional[Mapping[str, str]] = None, **kwargs) -> 'TagFileOutput':
        """
        get tag file input object
        :param workdir: local work directory
        :param pool: file pool object
        :param environ: environment variables
        :return: tag file input object
        """
        environ = _process_environ(environ)
        _tag = _check_pool_tag(env_template(self.__tag, environ))
        _local = os.path.normpath(
            os.path.abspath(os.path.join(workdir, _check_workdir_file(env_template(self.__local, environ)))))

        return TagFileOutput(
            pool=pool, local=_local, tag=_tag,
        )


class TagFileOutput(FileOutput, _ITagFileOutput):
    def __init__(self, pool: FilePool, local: str, tag: str):
        """
        :param pool: file pool
        :param local: local path
        :param tag: pool tag
        """
        self.__pool = pool
        self.__local = local
        self.__tag = tag

        _ITagFileOutput.__init__(self, self.__local, self.__tag)

    @property
    def tag(self) -> str:
        return self.__tag

    @property
    def local(self) -> str:
        return self.__local

    def __call__(self, output_start: Optional[Callable[['TagFileOutput'], None]] = None,
                 output_complete: Optional[Callable[['TagFileOutput'], None]] = None, **kwargs):
        """
        execute this file output
        """
        wrap_empty(output_start)(self)
        self.__pool[self.__tag] = self.__local
        wrap_empty(output_complete)(self)



from typing import Union, List, Optional, Mapping, Callable

from hbutils.string import env_template

from .base import _ICommandBase, CommandMode
from ...control.model import ResourceLimit, Identification, RunResult
from ...control.run import common_run, timing_run, mutual_run
from ...utils import eclosing, wrap_empty


class Command(_ICommandBase):
    def __init__(self, args: Union[str, List[str]], shell: bool,
                 workdir: Optional[str], environ: Mapping[str, str],
                 identification: Identification, resources: ResourceLimit,
                 mode, stdin, stdout, stderr):
        """
        :param args: arguments
        :param shell: use shell mode
        :param workdir: work directory
        :param environ: environment variables
        :param identification: identification used
        :param resources: resource limits
        :param mode: command mode value
        :param stdin: stdin file
        :param stdout: stdout file
        :param stderr: stderr file
        """
        self.__args = args
        self.__shell = shell

        self.__environ = environ
        self.__workdir = env_template(workdir, self.__environ)

        self.__identification = identification
        self.__resources = resources

        self.__mode = mode
        self.__stdin = env_template(stdin, self.__environ) if isinstance(stdin, str) else stdin
        self.__stdout = env_template(stdout, self.__environ) if isinstance(stdout, str) else stdout
        self.__stderr = env_template(stderr, self.__environ) if isinstance(stderr, str) else stderr

        _ICommandBase.__init__(self, self.__args, self.__shell, self.__workdir,
                               self.__identification, self.__resources, self.__mode)

    @property
    def args(self) -> Union[str, List[str]]:
        return self.__args

    @property
    def shell(self) -> bool:
        return self.__shell

    @property
    def environ(self) -> Mapping[str, str]:
        return self.__environ

    @property
    def workdir(self) -> str:
        return self.__workdir

    @property
    def identification(self) -> Identification:
        return self.__identification

    @property
    def resources(self) -> ResourceLimit:
        return self.__resources

    @property
    def mode(self) -> CommandMode:
        return self.__mode

    @property
    def stdin(self):
        return self.__stdin

    @property
    def stdout(self):
        return self.__stdout

    @property
    def stderr(self):
        return self.__stderr

    __RUN_FUNCTION = {
        CommandMode.COMMON: common_run,
        CommandMode.TIMING: timing_run,
        CommandMode.MUTUAL: mutual_run,
    }

    def __call__(self, command_start: Optional[Callable[['Command'], None]] = None,
                 command_complete: Optional[Callable[['Command', RunResult], None]] = None, **kwargs) -> RunResult:
        """
        run this command
        :return: run result
        """
        wrap_empty(command_start)(self)
        if isinstance(self.__stdin, str) and self.__mode != CommandMode.MUTUAL:
            stdin = open(self.__stdin, 'rb', 0)
            stdin_need_close = True
        else:
            stdin = self.__stdin
            stdin_need_close = False

        if isinstance(self.__stdout, str):
            stdout = open(self.__stdout, 'wb', 0)
            stdout_need_close = True
        else:
            stdout = self.__stdout
            stdout_need_close = False

        if isinstance(self.__stderr, str):
            stderr = open(self.__stderr, 'wb', 0)
            stderr_need_close = True
        else:
            stderr = self.__stderr
            stderr_need_close = False

        with eclosing(stdin, stdin_need_close) as fstdin, \
                eclosing(stdout, stdout_need_close) as fstdout, \
                eclosing(stderr, stderr_need_close) as fstderr:
            _result = self.__RUN_FUNCTION[self.__mode](
                args=self.__args, shell=self.__shell,
                stdin=fstdin, stdout=fstdout, stderr=fstderr,
                environ=self.__environ, cwd=self.__workdir,
                resources=self.__resources, identification=self.__identification,
            )

        wrap_empty(command_complete)(self, _result)
        return _result



from enum import unique, IntEnum
from typing import Mapping

from .base import SectionInfoTemplate
from .local import LocalSectionInfoTemplate
from .static import StaticSectionInfoTemplate
from .tag import TagSectionInfoTemplate


@unique
class SectionInfoType(IntEnum):
    LOCAL = 1
    TAG = 2
    STATIC = 3

    @classmethod
    def loads(cls, value) -> 'SectionInfoType':
        """
        Load SectionInfoType from value
        :param value: raw value
        :return: info info type object
        """
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            if value.upper() in cls.__members__.keys():
                return cls.__members__[value.upper()]
            else:
                raise KeyError('Unknown info info type - {actual}.'.format(actual=repr(value)))
        elif isinstance(value, int):
            _mapping = {v.value: v for k, v in cls.__members__.items()}
            if value in _mapping.keys():
                return _mapping[value]
            else:
                raise ValueError('Unknown info info type value - {actual}'.format(actual=repr(value)))
        else:
            raise TypeError('Int, str or {cls} expected but {actual} found.'.format(
                cls=cls.__name__,
                actual=repr(type(value).__name__)
            ))


# noinspection DuplicatedCode
_TYPE_TO_TEMPLATE_CLASS = {
    SectionInfoType.LOCAL: LocalSectionInfoTemplate,
    SectionInfoType.TAG: TagSectionInfoTemplate,
    SectionInfoType.STATIC: StaticSectionInfoTemplate,
}


def _load_info_template_from_json(json: Mapping[str, str]) -> SectionInfoTemplate:
    """
    load template object from json data
    :param json: json data
    :return: info info template object
    """
    if 'type' not in json.keys():
        raise KeyError('Key {type} not found.'.format(type=repr('type')))

    _type = SectionInfoType.loads(json['type'])
    _json = dict(json)
    del _json['type']

    return _TYPE_TO_TEMPLATE_CLASS[_type](**_json)


def _load_info_template_from_string(string: str) -> SectionInfoTemplate:
    """
    load template object from string (should split with :)
    :param string: string value
    :return: file info template object
    """
    _items = string.split(':')
    _type, _args = SectionInfoType.loads(_items[0].strip()), [_item.strip() for _item in _items[1:]]
    return _TYPE_TO_TEMPLATE_CLASS[_type](*_args)


def load_info_template(data) -> SectionInfoTemplate:
    """
    load info info template object from data
    :param data: raw data
    :return: info info template object
    """
    if isinstance(data, SectionInfoTemplate):
        return data
    elif isinstance(data, dict):
        return _load_info_template_from_json(data)
    elif isinstance(data, str):
        return _load_info_template_from_string(data)
    else:
        raise TypeError('Json, string or {type} expected but {actual} found.'.format(
            type=SectionInfoTemplate.__name__, actual=repr(type(data).__name__)))



import os
import pickle
from multiprocessing import Event, Value, Lock
from multiprocessing.synchronize import Event as EventClass
from threading import Thread
from typing import Optional, Tuple, Mapping

from .base import measure_thread, killer_thread, read_all_from_bytes_stream, GeneralProcess
from .decorator import process_setter
from .executor import get_child_executor_func
from ..model import ResourceLimit
from ...utils import ValueProxy


class CommonProcess(GeneralProcess):
    def __init__(self, start_time: float,
                 communicate_event: EventClass, communicate_complete: EventClass,
                 communicate_stdin: ValueProxy, communicate_stdout: ValueProxy, communicate_stderr: ValueProxy,
                 resources: ResourceLimit, process_result_func, lifetime_event: EventClass):
        self.__lock = Lock()
        GeneralProcess.__init__(self, start_time, resources, process_result_func, lifetime_event, self.__lock)

        self.__communicate_event = communicate_event
        self.__communicate_complete = communicate_complete
        self.__communicate_stdin = communicate_stdin
        self.__communicate_stdout = communicate_stdout
        self.__communicate_stderr = communicate_stderr
        self.__communicated = False

    def __communicate(self, stdin: Optional[bytes] = None, wait: bool = True) -> Optional[Tuple[bytes, bytes]]:
        if not self.__communicated:
            self.__communicate_stdin.value = stdin or b''
            self.__communicate_event.set()
            self.__communicated = True

            if wait:
                self.__communicate_complete.wait()
                return self.__communicate_stdout.value, self.__communicate_stderr.value
            else:
                return None
        else:
            raise RuntimeError('Already communicated.')

    def __exit(self):
        if not self.__communicated:
            self.__communicate()
        self._wait_for_end()

    def communicate(self, stdin: Optional[bytes] = None, wait: bool = True) -> Optional[Tuple[bytes, bytes]]:
        with self.__lock:
            return self.__communicate(stdin, wait)

    @property
    def stdin(self) -> Optional[bytes]:
        with self.__lock:
            return self.__communicate_stdin.value

    @property
    def stdout(self) -> Optional[bytes]:
        with self.__lock:
            return self.__communicate_stdout.value

    @property
    def stderr(self) -> Optional[bytes]:
        with self.__lock:
            return self.__communicate_stderr.value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.__lock:
            self.__exit()


# Attention: only real_time_limit will be processed in this function, other limits will be processed in decorator
# noinspection DuplicatedCode,PyIncorrectDocstring,PyUnresolvedReferences
@process_setter
def common_process(args, preexec_fn=None, resources=None,
                   environ: Optional[Mapping[str, str]] = None) -> CommonProcess:
    """
    Create an common process
    :param args: arguments for execution
    :param shell: use shell to execute args
    :param preexec_fn: pre execute function to attach before that
    :param resources: resource limit
    :param environ: environment variables
    :param cwd: new work dir
    :param identification: user and group for execution
    :return: CommonProcess object to do run
    """
    resources = ResourceLimit.loads(resources)
    _full_lifetime_complete = Event()
    environ = dict(environ or {})

    _executor_prepare_ok = Event()
    _exception_get, _exception_put = os.pipe()

    _parent_initialized = Event()
    _start_time = Value('d', 0.0)
    _start_time_ok = Event()

    # noinspection DuplicatedCode
    def _execute_parent() -> CommonProcess:
        os.close(stdin_read)
        os.close(stdout_write)
        os.close(stderr_write)

        # measure thread
        _measure_thread, _measure_initialized, _process_complete, _measure_complete, _result_proxy = measure_thread(
            start_time_ok=_start_time_ok,
            start_time=_start_time,
            child_pid=child_pid,
        )

        # killer thread
        _killer_thread, _killer_initialized = killer_thread(
            start_time_ok=_start_time_ok,
            start_time=_start_time,
            child_pid=child_pid,
            real_time_limit=resources.max_real_time,
            process_complete=_process_complete,
        )

        # communication thread
        _communicate_initialized, _communicate_event, _communicate_complete = Event(), Event(), Event()
        _communicate_stdin, _communicate_stdout, _communicate_stderr = ValueProxy(), ValueProxy(), ValueProxy()

        def _communicate_func():
            with os.fdopen(stdin_write, 'wb', 0) as fstdin, \
                    os.fdopen(stdout_read, 'rb', 0) as fstdout, \
                    os.fdopen(stderr_read, 'rb', 0) as fstderr:
                _communicate_initialized.set()
                _communicate_event.wait()

                _stdout, _stderr = None, None

                def _read_stdout():
                    nonlocal _stdout
                    _stdout = read_all_from_bytes_stream(fstdout)

                def _read_stderr():
                    nonlocal _stderr
                    _stderr = read_all_from_bytes_stream(fstderr)

                def _write_stdin():
                    try:
                        fstdin.write(_communicate_stdin.value)
                        fstdin.flush()
                    except BrokenPipeError:
                        pass
                    finally:
                        fstdin.close()

                _stdin_thread = Thread(target=_write_stdin)
                _stdout_thread = Thread(target=_read_stdout)
                _stderr_thread = Thread(target=_read_stderr)

                # write stdin into stream
                _stdin_thread.start()
                _stdout_thread.start()
                _stderr_thread.start()

                # waiting for receiving of stdout and stderr
                _stdin_thread.join()
                _stdout_thread.join()
                _stderr_thread.join()

                # set stderr
                _communicate_stdout.value = _stdout
                _communicate_stderr.value = _stderr
                _communicate_complete.set()

                # ending of all the process
                _process_complete.wait()
                _measure_complete.wait()
                _full_lifetime_complete.set()

        _communicate_thread = Thread(target=_communicate_func)

        # waiting for prepare ok
        _executor_prepare_ok.wait()
        with os.fdopen(_exception_get, 'rb', 0) as ef:
            _exception = pickle.load(ef)
            if _exception:
                raise _exception

        # start all the threads and services
        _measure_thread.start()
        _killer_thread.start()
        _communicate_thread.start()

        # wait for all the thread initialized
        _measure_initialized.wait()
        _killer_initialized.wait()
        _communicate_initialized.wait()
        _parent_initialized.set()

        _start_time_ok.wait()
        return CommonProcess(
            start_time=_start_time.value,
            communicate_event=_communicate_event,
            communicate_complete=_communicate_complete,
            communicate_stdin=_communicate_stdin,
            communicate_stdout=_communicate_stdout,
            communicate_stderr=_communicate_stderr,
            resources=resources,
            process_result_func=lambda: _result_proxy.value,
            lifetime_event=_full_lifetime_complete,
        )

    stdin_read, stdin_write = os.pipe()
    stdout_read, stdout_write = os.pipe()
    stderr_read, stderr_write = os.pipe()

    _execute_child = get_child_executor_func(
        args, dict(environ or {}), preexec_fn,
        _executor_prepare_ok, (_exception_get, _exception_put),
        _parent_initialized,
        _start_time_ok, _start_time,
        (stdin_read, stdin_write),
        (stdout_read, stdout_write),
        (stderr_read, stderr_write),
    )

    child_pid = os.fork()

    if not child_pid:
        _execute_child()
    else:
        return _execute_parent()



from contextlib import contextmanager
from typing import TypeVar

_T = TypeVar('_T')


@contextmanager
def eclosing(obj: _T, close: bool = True) -> _T:
    try:
        yield obj
    finally:
        if close:
            obj.close()



import os
from functools import partial
from typing import Mapping, Optional

from hbutils.string import env_template

from .base import _ISection, _check_section_name, ENV_PJI_SECTION_NAME
from .section import Section
from ..info import SectionInfoMappingTemplate
from ..input import FileInputCollectionTemplate
from ..output import FileOutputCollectionTemplate
from ...base import _process_environ, _check_os_path
from ...command import CommandCollectionTemplate
from ....control.model import Identification, ResourceLimit
from ....utils import FilePool


class SectionTemplate(_ISection):
    def __init__(self, name: str, commands,
                 identification=None, resources=None, environ=None,
                 inputs=None, outputs=None, infos=None, info_dump=None):
        """
        :param name: section name
        :param commands: commands
        :param identification: identification
        :param resources: resource limits
        :param environ: environment variables (${ENV} supported)
        :param inputs: input collection template
        :param outputs: output collection template
        :param infos: information collection template
        :param info_dump: information result dump
        """
        self.__name = name
        self.__commands = CommandCollectionTemplate.loads(commands)

        self.__identification = Identification.loads(identification)
        self.__resources = ResourceLimit.loads(resources)
        self.__environ = _process_environ(environ)

        self.__inputs = FileInputCollectionTemplate.loads(inputs)
        self.__outputs = FileOutputCollectionTemplate.loads(outputs)
        self.__infos = SectionInfoMappingTemplate.loads(infos)
        self.__info_dump = info_dump

        _ISection.__init__(self, self.__name, self.__identification, self.__resources,
                           self.__environ, self.__inputs, self.__outputs, self.__infos, self.__commands)

    @property
    def name(self) -> str:
        return self.__name

    @property
    def commands(self) -> CommandCollectionTemplate:
        return self.__commands

    @property
    def identification(self) -> Identification:
        return self.__identification

    @property
    def resources(self) -> ResourceLimit:
        return self.__resources

    @property
    def environ(self) -> Mapping[str, str]:
        return self.__environ

    @property
    def inputs(self) -> FileInputCollectionTemplate:
        return self.__inputs

    @property
    def outputs(self) -> FileOutputCollectionTemplate:
        return self.__outputs

    @property
    def infos(self) -> SectionInfoMappingTemplate:
        return self.__infos

    @property
    def info_dump(self) -> str:
        return self.__info_dump

    def __call__(self, scriptdir: str, pool: Optional[FilePool] = None,
                 identification=None, resources=None, environ=None, **kwargs) -> Section:
        """
        generate section object
        :param scriptdir: script directory
        :param pool: file pool
        :param identification: identification
        :param resources: resource limits
        :param environ: environment variables
        :param kwargs: any other arguments
        :return: section object
        """
        if 'workdir' in kwargs.keys():
            raise KeyError('Workdir is not allowed to pass into section template.')

        environ = dict(_process_environ(self.__environ, environ, enable_ext=True))
        _name = _check_section_name(env_template(self.__name, environ))
        environ[ENV_PJI_SECTION_NAME] = _name

        _identification = Identification.merge(Identification.loads(identification), self.__identification)
        _resources = ResourceLimit.merge(ResourceLimit.loads(resources), self.__resources)
        _info_dump = os.path.normpath(
            os.path.abspath(os.path.join(scriptdir, _check_os_path(
                env_template(self.__info_dump, environ))))) if self.__info_dump else None

        arguments = dict(kwargs)
        arguments.update(
            scriptdir=scriptdir,
            pool=pool,
            environ=environ,
            identification=_identification,
            resources=_resources,
        )

        return Section(
            name=_name,
            identification=_identification, resources=_resources, environ=environ,
            commands_getter=partial(self.__commands, **arguments),
            inputs_getter=partial(self.__inputs, **arguments),
            outputs_getter=partial(self.__outputs, **arguments),
            infos_getter=partial(self.__infos, **arguments),
            info_dump=_info_dump,
        )

    @classmethod
    def loads(cls, data) -> 'SectionTemplate':
        """
        load section template from data
        :param data: raw data
        :return: section template object
        """
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls(**data)
        else:
            raise TypeError('Json or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))



import os
import re
import shutil
import tempfile
from multiprocessing import Lock
from typing import Mapping, Optional

from pysyslimit import FilePermission, SystemUser, SystemGroup, chown, chmod

from .path import makedirs


def _auto_delete(filename: str):
    if os.path.isdir(filename):
        shutil.rmtree(filename, ignore_errors=True)
    else:
        os.remove(filename)


def _check_from_file(from_file: str):
    if not os.path.exists(from_file):
        raise FileNotFoundError('File {filename} not found.'.format(filename=repr(from_file)))
    if not os.access(from_file, os.R_OK):
        raise PermissionError('File {filename} unreadable.'.format(filename=repr(from_file)))


def _prepare_for_to_file(to_file: str, privilege=None, user=None, group=None):
    if os.path.exists(to_file):
        _auto_delete(to_file)
    _parent_path, _ = os.path.split(to_file)
    makedirs(_parent_path, privilege, user, group)


def auto_copy_file(from_file: str, to_file: str, privilege=None, user=None, group=None):
    privilege = FilePermission.loads(privilege) if privilege else None
    user = SystemUser.loads(user) if user else None
    group = SystemGroup.loads(group) if group else None

    _check_from_file(from_file)
    _prepare_for_to_file(to_file, privilege, user, group)
    if os.path.isdir(from_file):
        shutil.copytree(from_file, to_file)
    else:
        shutil.copyfile(from_file, to_file, follow_symlinks=True)

    if privilege:
        chmod(to_file, privilege, recursive=True)
    if user or group:
        chown(to_file, user, group, recursive=True)


def auto_link_file(from_file: str, to_file: str):
    _check_from_file(from_file)
    _prepare_for_to_file(to_file)
    os.symlink(from_file, to_file)


class FilePool:
    __FILE_NAME = 'file'
    __TAG_NAME_REGEXP = r'[a-zA-Z0-9_]+'

    def __init__(self, init: Optional[Mapping[str, str]] = None):
        self.__root_dir = tempfile.TemporaryDirectory()
        self.__file_dirs = {}

        self.__lock = Lock()
        self.__init_pool(init or {})

    def __init_pool(self, init: Mapping[str, str]):
        try:
            for tag, fdir in init.items():
                self.check_tag_name(tag)
                self.__create_tag_file(tag, fdir)
        except Exception as err:
            self.__close()
            raise err

    @classmethod
    def check_tag_name(cls, tag: str):
        if not re.fullmatch(cls.__TAG_NAME_REGEXP, tag):
            raise KeyError(
                'Tag name should only contains a-z, A-Z, 0-9 and _, but {actual} found.'.format(actual=repr(tag)))

    def __check_tag_exist(self, tag: str):
        if tag not in self.__file_dirs.keys():
            raise KeyError('Tag {tag} not exist.'.format(tag=repr(tag)))

    def __get_tag_file(self, tag: str) -> str:
        return os.path.join(self.__file_dirs[tag].name, self.__FILE_NAME)

    def __create_tag_file(self, tag: str, filename: str):
        self.__file_dirs[tag] = tempfile.TemporaryDirectory(dir=self.__root_dir.name)
        auto_copy_file(filename, self.__get_tag_file(tag))

    def __remove_tag_file(self, tag: str):
        self.__file_dirs[tag].cleanup()
        del self.__file_dirs[tag]

    def __export_tag_file(self, tag: str, filename: str, privilege=None, user=None, group=None):
        auto_copy_file(self.__get_tag_file(tag), filename, privilege, user, group)

    def __link_tag_file(self, tag: str, filename: str):
        auto_link_file(self.__get_tag_file(tag), filename)

    def __clear(self):
        try:
            for tag, fdir in self.__file_dirs.items():
                if os.path.exists(fdir.name):
                    fdir.cleanup()
        finally:
            for tag, fdir in list(self.__file_dirs.items()):
                if not os.path.exists(fdir.name):
                    del self.__file_dirs[tag]

    def __close(self):
        self.__clear()
        if os.path.exists(self.__root_dir.name):
            self.__root_dir.cleanup()

    def __getitem__(self, tag: str):
        with self.__lock:
            self.check_tag_name(tag)
            self.__check_tag_exist(tag)
            return self.__get_tag_file(tag)

    def __setitem__(self, tag, filename):
        with self.__lock:
            self.check_tag_name(tag)
            if tag in self.__file_dirs.keys():
                self.__remove_tag_file(tag)
            return self.__create_tag_file(tag, filename)

    def __delitem__(self, tag):
        with self.__lock:
            self.check_tag_name(tag)
            self.__check_tag_exist(tag)
            return self.__remove_tag_file(tag)

    def __contains__(self, tag):
        with self.__lock:
            return tag in self.__file_dirs

    def export(self, tag: str, filename: str, privilege=None, user=None, group=None):
        with self.__lock:
            return self.__export_tag_file(tag, filename, privilege, user, group)

    def link(self, tag: str, filename: str):
        with self.__lock:
            return self.__link_tag_file(tag, filename)

    def clear(self):
        with self.__lock:
            return self.__clear()

    def close(self):
        with self.__lock:
            return self.__close()

    def __enter__(self):
        with self.__lock:
            return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.__lock:
            self.__close()



import codecs
import os
import warnings
from abc import ABCMeta
from typing import Optional, Mapping, Tuple

from hbutils.string import env_template

from .base import SectionInfoTemplate, SectionInfo
from ...base import _check_pool_tag, _check_workdir_file, _process_environ
from ....utils import get_repr_info, FilePool


class _ITagSectionInfo(metaclass=ABCMeta):
    def __init__(self, tag: str, file: Optional[str]):
        """        
        :param tag: file pool tag 
        :param file: sub file of dir
        """
        self.__tag = tag
        self.__file = file

    def __repr__(self):
        """
        :return: representation string 
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('tag', lambda: repr(self.__tag)),
                ('file', lambda: repr(self.__file), lambda: self.__file is not None)
            ]
        )


class TagSectionInfoTemplate(SectionInfoTemplate, _ITagSectionInfo):
    def __init__(self, tag: str, file: Optional[str] = None):
        """        
        :param tag: file pool tag 
        :param file: sub file of dir
        """
        self.__tag = tag
        self.__file = file

        _ITagSectionInfo.__init__(self, self.__tag, self.__file)

    @property
    def tag(self) -> str:
        return self.__tag

    @property
    def file(self) -> str:
        return self.__file

    def __call__(self, pool: FilePool, environ: Optional[Mapping[str, str]] = None, **kwargs) -> 'TagSectionInfo':
        """
        get tag info info object
        :param pool: file pool object
        :param environ: environment variables
        :return: tag info info object
        """
        environ = _process_environ(environ)
        _tag = _check_pool_tag(env_template(self.__tag, environ))
        if self.__file is not None:
            _file = os.path.normpath(_check_workdir_file(env_template(self.__file, environ)))
        else:
            _file = None

        return TagSectionInfo(pool=pool, tag=_tag, file=_file)


class TagSectionInfo(SectionInfo, _ITagSectionInfo):
    def __init__(self, pool: FilePool, tag: str, file: Optional[str] = None):
        """
        :param pool: file pool
        :param tag: file pool tag 
        :param file: sub file of dir
        """
        self.__pool = pool
        self.__tag = tag
        self.__file = file

        _ITagSectionInfo.__init__(self, self.__tag, self.__file)

    @property
    def tag(self) -> str:
        return self.__tag

    @property
    def file(self) -> str:
        return self.__file

    def __call__(self, **kwargs) -> Tuple[bool, Optional[str]]:
        """
        execute this info info
        """

        def _result_func():
            _tagged_file = self.__pool[self.__tag]
            if os.path.isdir(_tagged_file):
                if self.__file is None:
                    raise RuntimeError('Tag {tag} represent a dir but file is empty.'.format(tag=repr(self.__tag)))
                else:
                    _output_file = os.path.join(_tagged_file, self.__file)
            else:
                if self.__file is not None:
                    warnings.warn(RuntimeWarning(
                        'Tag {tag} represent a file, {file} data item will be ignored.'.format(tag=repr(self.__tag),
                                                                                               file=repr('file'))))
                _output_file = _tagged_file

            with codecs.open(_output_file, 'r') as f:
                return f.read()

        try:
            return True, _result_func()
        except (KeyError, RuntimeError):
            return False, None



import codecs
import os
from abc import ABCMeta
from typing import Optional, Mapping, Tuple

from hbutils.string import env_template

from .base import SectionInfoTemplate, SectionInfo
from ...base import _check_workdir_file, _process_environ
from ....utils import get_repr_info


class _ILocalSectionInfo(metaclass=ABCMeta):
    def __init__(self, file: str):
        """
        :param file: local path
        """
        self.__file = file

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('file', lambda: repr(self.__file)),
            ]
        )


class LocalSectionInfoTemplate(SectionInfoTemplate, _ILocalSectionInfo):
    def __init__(self, file: str):
        """
        :param file: local path
        """
        self.__file = file

        _ILocalSectionInfo.__init__(self, self.__file)

    @property
    def file(self) -> str:
        return self.__file

    def __call__(self, workdir: str, environ: Optional[Mapping[str, str]] = None, **kwargs) -> 'LocalSectionInfo':
        """
        generate local info info object from extension information
        :param workdir: work directory
        :param environ: environment variable
        :return: local info info object
        """
        environ = _process_environ(environ)
        _local = os.path.normpath(
            os.path.abspath(os.path.join(workdir, _check_workdir_file(env_template(self.__file, environ)))))

        return LocalSectionInfo(file=_local)


class LocalSectionInfo(SectionInfo, _ILocalSectionInfo):
    def __init__(self, file: str):
        """
        :param file: local path
        """
        self.__file = file

        _ILocalSectionInfo.__init__(self, self.__file)

    @property
    def file(self) -> str:
        return self.__file

    def __call__(self, **kwargs) -> Tuple[bool, Optional[str]]:
        """
        execute this info info
        """

        def _result_func():
            if os.path.isdir(self.__file):
                raise IsADirectoryError('Path {path} is directory.'.format(path=repr(self.__file)))
            with codecs.open(self.__file, 'r') as file:
                return file.read()

        try:
            return True, _result_func()
        except (FileNotFoundError, IsADirectoryError, PermissionError):
            return False, None



from ...utils import auto_decode_support, auto_encode_support

_auto_encode = auto_encode_support(lambda x: x)
_auto_decode = auto_decode_support(lambda x: x)


def _try_write(stream, data):
    _bytes_data = _auto_encode(data)
    _str_data = _auto_decode(_bytes_data)

    try:
        return stream.write(_bytes_data)
    except TypeError:
        return stream.write(_str_data)


def _try_read_to_bytes(stream):
    return _auto_encode(stream.read())



from .base import SectionInfoTemplate, SectionInfo
from .general import SectionInfoType, load_info_template
from .local import LocalSectionInfoTemplate, LocalSectionInfo
from .mapping import SectionInfoMappingTemplate, SectionInfoMapping
from .static import StaticSectionInfoTemplate, StaticSectionInfo
from .tag import TagSectionInfoTemplate, TagSectionInfo
